"""
Quantiphyse - Widgets for data smoothing

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import QpWidget, TitleWidget, OverlayCombo, NumericOption
from quantiphyse.utils import QpException

from .process import SmoothingProcess
    
class SmoothingWidget(QpWidget):
    """
    Gaussian smoothing widget
    """
    def __init__(self, **kwargs):
        super(SmoothingWidget, self).__init__(name="Smoothing", icon="smooth.png", 
                                              desc="Gaussian smoothing", group="Processing", **kwargs)
        
    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self, title="Data Smoothing", subtitle="Smooth data using a Gaussian kernel", help="smoothing")
        vbox.addWidget(title)

        hbox = QtGui.QHBoxLayout()
        gbox = QtGui.QGroupBox()
        gbox.setTitle("Options")
        grid = QtGui.QGridLayout()
        gbox.setLayout(grid)

        grid.addWidget(QtGui.QLabel("Data to smooth"), 0, 0)
        self.data_combo = OverlayCombo(self.ivm)
        self.data_combo.currentIndexChanged.connect(self.data_changed)
        grid.addWidget(self.data_combo, 0, 1)
        self.sigma = NumericOption("Sigma (mm)", grid, xpos=0, ypos=1, minval=0, step=0.1, default=1.0)
        grid.addWidget(QtGui.QLabel("Output name"), 2, 0)
        self.output_name = QtGui.QLineEdit()
        grid.addWidget(self.output_name, 2, 1)

        hbox.addWidget(gbox)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        hbox = QtGui.QHBoxLayout()
        self.run_btn = QtGui.QPushButton("Run")
        self.run_btn.clicked.connect(self.run)
        hbox.addWidget(self.run_btn)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        vbox.addStretch(1) 
        self.data_changed()  
    
    def data_changed(self):
        self.output_name.setText("%s_smoothed" % self.data_combo.currentText())
        self.run_btn.setEnabled(self.data_combo.currentText() in self.ivm.data)

    def batch_options(self):
        return "Smooth", {"data" : self.data_combo.currentText(),
                          "sigma" : self.sigma.spin.value(),
                          "output-name" : self.output_name.text()}

    def run(self):
        process = SmoothingProcess(self.ivm)
        process.run(self.batch_options()[1])



"""
Quantiphyse - Basic data processes

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .process import Process
from .feat_pca import PcaFeatReduce as PCA
from . import normalisation

__all__ = ["Process", "PCA", "normalisation"]



"""
Quantiphyse - Analysis processes for data simulation

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import math

import numpy as np
import scipy.ndimage.interpolation

from quantiphyse.data import DataGrid
from quantiphyse.utils import QpException
from quantiphyse.processes import Process

class AddNoiseProcess(Process):
    """
    Simple process for adding gaussian noise
    """
    PROCESS_NAME = "AddNoise"

    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        data = self.get_data(options)

        output_name = options.pop("output-name", "%s_noisy" % data.name)
        if "std" in options:
            std = float(options.pop("std"))
        elif "percent" in options:
            percent = float(options.pop("percent"))
            std = np.mean(data.raw()) * float(percent) / 100
        elif "snr" in options:
            snr = float(options.pop("snr"))
            roi = self.get_roi(options, grid=data.grid)
            signal = np.mean(data.raw()[roi.raw() > 0])
            std = signal / snr
        else:
            raise QpException("AddNoiseProcess: Must specify either std, percent or snr")

        noise = np.random.normal(loc=0, scale=std, size=list(data.grid.shape) + [data.nvols,])
        if data.nvols == 1: 
            noise = np.squeeze(noise, -1)
        noisy_data = data.raw() + noise
        self.ivm.add(noisy_data, grid=data.grid, name=output_name, make_current=True)

class SimMotionProcess(Process):
    """
    Simple process for adding gaussian noise
    """
    PROCESS_NAME = "SimMotion"

    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        data = self.get_data(options)
        if data.ndim != 4:
            raise QpException("Can only simulate motion on 4D data")

        output_name = options.pop("output-name", "%s_moving" % data.name)
        std = float(options.pop("std", "0"))
        std_voxels = [std / size for size in data.grid.spacing]
        std_degrees = float(options.pop("std_rot", "0"))
        order = int(options.pop("order", "1"))
        output_grid = data.grid
        output_shape = data.grid.shape
            
        padding = options.pop("padding", 0)
        if padding > 0:
            padding_voxels = [int(math.ceil(padding / size)) for size in data.grid.spacing]
            for dim in range(3):
                if data.shape[dim] == 1:
                    padding_voxels[dim] = 0
            # Need to adjust the origin so the output data lines up with the input
            output_origin = np.copy(data.grid.origin)
            output_shape = np.copy(data.grid.shape)
            output_affine = np.copy(data.grid.affine)
            for axis in range(3):
                output_origin[axis] -= np.dot(padding_voxels, data.grid.transform[axis, :])
                output_shape[axis] += 2*padding_voxels[axis]
            output_affine[:3, 3] = output_origin
            output_grid = DataGrid(output_shape, output_affine)

        moving_data = np.zeros(list(output_shape) + [data.nvols,])
        centre_offset = output_shape / 2
        for vol in range(data.nvols):
            voldata = data.volume(vol)
            if padding > 0:
                voldata = np.pad(voldata, [(v, v) for v in padding_voxels], 'constant', constant_values=0) 
            shift = np.random.normal(scale=std_voxels, size=3)
            for dim in range(3):
                if voldata.shape[dim] == 1:
                    shift[dim] = 0
            shifted_data = scipy.ndimage.shift(voldata, shift, order=order)

            # Generate random rotation and scale it to the random angle
            required_angle = np.random.normal(scale=std_degrees, size=1)
            rot = scipy.spatial.transform.Rotation.random().as_rotvec()
            rot_angle = np.degrees(np.sqrt(np.sum(np.square(rot))))
            rot *= required_angle / rot_angle
            rot_matrix = scipy.spatial.transform.Rotation.from_rotvec(rot).as_matrix()

            offset=centre_offset-centre_offset.dot(rot_matrix)
            rotated_data = scipy.ndimage.affine_transform(shifted_data, rot_matrix.T, offset=offset, order=order)
            moving_data[..., vol] = rotated_data

        self.ivm.add(moving_data, grid=output_grid, name=output_name, make_current=True)



"""
Quantiphyse - Analysis process tests

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import unittest

import numpy as np
import pandas as pd

from quantiphyse.processes import Process
from quantiphyse.test import ProcessTest

class AnalysisProcessTest(ProcessTest):
    
    def testCalcVolumes(self):
        yaml = """       
  - CalcVolumes:
        roi: mask
        output-name: mask_vol

  - SaveExtras: 
        mask_vol: mask_vol.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("mask_vol" in self.ivm.extras)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "mask_vol.tsv")))

    def testSummaryStatsOldName(self):
        yaml = """
  - OverlayStats:
        data: data_3d
        output-name: testdata_stats

  - SaveExtras: 
        testdata_stats: testdata_stats.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_stats" in self.ivm.extras)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "testdata_stats.tsv")))

    def testSummaryStats(self):
        yaml = """
  - DataStatistics:
        data: data_3d
        output-name: testdata_stats

  - SaveExtras: 
        testdata_stats: testdata_stats.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_stats" in self.ivm.extras)

        fname = os.path.join(self.output_dir, "case", "testdata_stats.tsv")
        self.assertTrue(os.path.exists(fname))
        df = pd.read_csv(fname, sep='\t')
        data = df.values
        self.assertEquals(data.shape[0], 5)
        self.assertEquals(data.shape[1], 2)
        self.assertAlmostEquals(data[0, 1], np.mean(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[1, 1], np.median(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[2, 1], np.std(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[3, 1], np.min(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[4, 1], np.max(self.data_3d), delta=0.01)

    def testSummaryStatsMultiple(self):
        yaml = """
  - DataStatistics:
        data: [data_3d, data_4d]
        output-name: testdata_stats

  - SaveExtras: 
        testdata_stats: testdata_stats.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_stats" in self.ivm.extras)

        fname = os.path.join(self.output_dir, "case", "testdata_stats.tsv")
        self.assertTrue(os.path.exists(fname))
        df = pd.read_csv(fname, sep='\t')
        data = df.values
        self.assertEquals(data.shape[0], 5)
        self.assertEquals(data.shape[1], 3)
        self.assertAlmostEquals(data[0, 1], np.mean(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[1, 1], np.median(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[2, 1], np.std(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[3, 1], np.min(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[4, 1], np.max(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[0, 2], np.mean(self.data_4d), delta=0.01)
        self.assertAlmostEquals(data[1, 2], np.median(self.data_4d), delta=0.01)
        self.assertAlmostEquals(data[2, 2], np.std(self.data_4d), delta=0.01)
        self.assertAlmostEquals(data[3, 2], np.min(self.data_4d), delta=0.01)
        self.assertAlmostEquals(data[4, 2], np.max(self.data_4d), delta=0.01)

    def testSummaryStatsRoi(self):
        yaml = """
  - KMeans:
        data: data_3d
        n-clusters: 3
        output-name: clusters
        
  - DataStatistics:
        data: [data_3d, data_4d]
        roi: clusters
        output-name: testdata_stats

  - SaveExtras: 
        testdata_stats: testdata_stats.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_stats" in self.ivm.extras)

        fname = os.path.join(self.output_dir, "case", "testdata_stats.tsv")
        self.assertTrue(os.path.exists(fname))
        df = pd.read_csv(fname, sep='\t')
        data = df.values
        self.assertEquals(data.shape[0], 5)
        self.assertEquals(data.shape[1], 7)
        clusters = self.ivm.data["clusters"].raw()

        for roi_region in range(3):
            data_3d = self.data_3d[clusters == roi_region+1]
            data_4d = self.data_4d[clusters == roi_region+1]
            self.assertAlmostEquals(data[0, 1+roi_region], np.mean(data_3d), delta=0.01)
            self.assertAlmostEquals(data[1, 1+roi_region], np.median(data_3d), delta=0.01)
            self.assertAlmostEquals(data[2, 1+roi_region], np.std(data_3d), delta=0.01)
            self.assertAlmostEquals(data[3, 1+roi_region], np.min(data_3d), delta=0.01)
            self.assertAlmostEquals(data[4, 1+roi_region], np.max(data_3d), delta=0.01)
            self.assertAlmostEquals(data[0, 4+roi_region], np.mean(data_4d), delta=0.01)
            self.assertAlmostEquals(data[1, 4+roi_region], np.median(data_4d), delta=0.01)
            self.assertAlmostEquals(data[2, 4+roi_region], np.std(data_4d), delta=0.01)
            self.assertAlmostEquals(data[3, 4+roi_region], np.min(data_4d), delta=0.01)
            self.assertAlmostEquals(data[4, 4+roi_region], np.max(data_4d), delta=0.01)

    def testRadialProfile(self):
        yaml = """ 
  - RadialProfile:
        data: data_3d
        centre: 2, 2, 2
        output-name: testdata_rp

  - SaveExtras: 
        testdata_rp: testdata_rp.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_rp" in self.ivm.extras)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "testdata_rp.tsv")))

    def testHistogram(self):
        yaml = """
  - Histogram:
        data: data_3d
        bins: 10
        output-name: testdata_hist
        roi: mask

  - SaveExtras: 
        testdata_hist: testdata_hist.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_hist" in self.ivm.extras)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "testdata_hist.tsv")))

    def testExecMultiply(self):
        yaml = """
  - Exec:
      test1: data_3d * 3.14159265  
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("test1" in self.ivm.data)

    def testExecDemean(self):
        yaml = """
  - Exec:
      test1: data_3d - np.mean(data_3d)
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("test1" in self.ivm.data)

    def testExecSubtractVolume(self):
        yaml = """
  - Exec:
      test1: data_3d - data_4d[..., 0]
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("test1" in self.ivm.data)

if __name__ == '__main__':
    unittest.main()



"""
Quantiphyse - Base class for a registration method

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets
  
import numpy as np

from quantiphyse.data import NumpyData
from quantiphyse.utils import LogSource, QpException

class RegMethod(LogSource):
    """
    A registration method

    Methods should implement, at a minimum, the ``reg`` method
    Methods which take options should implement ``interface`` and ``options``
    Methods may implement ``moco`` if motion correction is handled differently
    """
    def __init__(self, name, ivm, display_name=None):
        LogSource.__init__(self)
        self.name = name
        self.ivm = ivm
        if display_name is not None:
            self.display_name = display_name
        else:
            self.display_name = name

    @classmethod
    def apply_transform(cls, reg_data, transform, options, queue):
        """
        Apply a previously calculated transformation to a data set

        :param reg_data: QpData containing data to apply the transform to.
        :param transform: A transformation object as returned by one of the registration methods.
                          This must be compatible with the registration class (i.e.
                          typically generated by the same registration class)
        :param options: Method options as dictionary
        :param queue: Queue object which method may put progress information on to. Progress 
                      should be given as a number between 0 and 1.
        :return Tuple of QpData containing transformed data and log output as a string

        Standard options:

        :param output-space: `ref` to output in the same space as the transformation reference space,
                             `reg` to output in the same space as the input data,
                             `transform` to apply the transformation to the data space rather than
                             resampling the data (only possible for affine transformations)
        :param interp-order: Interpolation order (0=nearest neighbour, 1=linear, 2=quadratic, 3=cubic)
        """
        raise NotImplementedError("Registration method has not implemented 'apply_transform'")

    @classmethod
    def reg_3d(cls, reg_data, ref_data, options, queue):
        """
        3D Registration

        :param reg_data: 3D QpData containing data to register.
        :param ref_data: 3D QpData containing reference data.
        :param options: Method options as dictionary
        :param queue: Queue object which method may put progress information on to. Progress 
                      should be given as a number between 0 and 1.

        :return Tuple of three items. 
        
                First, A QpData containing registered data

                Second, if options contains ``output-transform : True``, transformation found. 
                This is either a QpData containing warp images (typically 3) or an Extra
                containing a matrix transformation. If ``output-transform`` is not given, or
                is not supported, returns None instead.

                Third, log information from the registration as a string.
        """
        raise NotImplementedError("Registration method has not implemented 'reg'")

    @classmethod
    def reg_4d(cls, reg_data, ref_data, options, queue):
        """
        4D Registration

        The default implementation simply registers each volume of the data independently. However,
        implementations can supply their own more optimal implementation if appropriate

        :param reg_data: 4D QpData containing data to register.
        :param ref_data: 3D QpData containing reference data.
        :param options: Method options as dictionary
        :param queue: Queue object which method may put progress information on to. Progress 
                      should be given as a number between 0 and 1.

        :return Tuple of three items. 
        
                First, A QpData containing registered data

                Second, if options contains ``output-transform : True``, sequence of transformations
                found, one for each volume in ``reg_data``. Each is either a QpData object containing 
                a sequence of 3 warp images or an Extra object containing a transformation matrix
                If ``output-transform`` is not given or not supported, returns None instead.

                Third, log information from the registration as a string.
        """
        if reg_data.ndim != 4:
            raise QpException("reg_4d expected 4D data")
        
        if options.get("output-space", "ref") == "ref":
            output_space = ref_data
        else:
            output_space = reg_data
        out_data = np.zeros(list(output_space.grid.shape) + [reg_data.nvols])

        transforms = []
        log = "Default 4D registration using multiple 3d registrations\n"
        for vol in range(reg_data.shape[-1]):
            log += "Registering volume %i of %i\n" % (vol+1, reg_data.shape[-1])
            reg_vol = NumpyData(reg_data.volume(vol), grid=reg_data.grid, name="regvol")
            #self.debug("Vol %i of %i" % (vol+1, reg_data.shape[-1]))
            if vol == options.get("ignore-idx", -1):
                # Ignore this index (e.g. because it is the same as the ref volume)
                if options.get("output-space", "ref") != "reg":
                    raise QpException("Can't ignore an index unless the output space is the registration data")
                out_data[..., vol] = reg_vol.raw()
                transforms.append(None)
            else:
                #self.debug("Calling reg_3d", cls, cls.reg_3d)
                # We did not remove output-space from the options so regdata should
                # come back in the appropriate space
                regdata, transform, vol_log = cls.reg_3d(reg_vol, ref_data, options, queue)
                out_data[..., vol] = regdata.raw()
                transforms.append(transform)
                log += vol_log
            queue.put(float(vol)/reg_data.shape[-1])

        return NumpyData(out_data, grid=output_space.grid, name=reg_data.name), transforms, log

    @classmethod
    def moco(cls, moco_data, ref, options, queue):
        """
        Motion correction
        
        The default implementation uses the ``reg_4d`` function to perform motion correction
        as registration to a common reference, however this function can have a custom
        implementation specific to motion correction if required.
        
        :param moco_data: A single 4D QpData containing data to motion correct.
        :param ref: Either 3D QpData containing reference data, or integer giving 
                    the volume index of ``moco_data`` to use
        :param options: Method options as dictionary
        :param queue: Queue object which method may put progress information on to. Progress 
                      should be given as a number between 0 and 1.
        
        :return Tuple of three items. 
        
                First, motion corrected QpData in the same space as ``moco_data``
        
                Second, if options contains ``output-transform : True``, sequence of transformations
                found, one for each volume in ``reg_data``. Each is either a QpData object containing 
                a sequence of 3 warp images or an Extra object containing a transformation matrix
                If ``output-transform`` is not given or not supported, returns None instead.

                Third, log information from the registration as a string.
        """
        if moco_data.ndim != 4:
            raise QpException("Cannot motion correct 3D data")
        
        log = "Default MOCO implementation using multiple 3d registrations\n"
        if isinstance(ref, int):
            if ref >= moco_data.nvols:
                raise QpException("Reference volume index of %i, but data has only %i volumes" % (ref, moco_data.nvols))
            ref = moco_data.volume(ref)

        options["output-space"] = "reg"
        out_data, transforms, moco_log = cls.reg_4d(moco_data, ref, options, queue)
        log += moco_log
        return out_data, transforms, log

    def interface(self, generic_options):
        """
        :param generic_options: Dictionary of generic registration options - this is provided so
               the method can customize the options offered depending on the registration mode.

        :return: QtGui.QWidget() to allow options to be controlled
        """
        return QtGui.QWidget()
        
    def options(self):
        """
        :return: Dictionary of options currently selected
        """ 
        return {}



"""
Quantiphyse - Experimental widgets (not currently in GUI)

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function, division

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import QpWidget, HelpButton

class ImageExportWidget(QpWidget):

    def __init__(self, **kwargs):
        super(ImageExportWidget, self).__init__(name="Image Export", desc="Export images and animations", icon="image_export", **kwargs)

    def init_ui(self):
        main_vbox = QtGui.QVBoxLayout()
        
        hbox = QtGui.QHBoxLayout()
        hbox.addWidget(QtGui.QLabel('<font size="5">Export animation</font>'))
        hbox.addStretch(1)
        hbox.addWidget(HelpButton(self))
        main_vbox.addLayout(hbox)

        hbox = QtGui.QHBoxLayout()
        b1 = QtGui.QPushButton('Generate', self)
        b1.clicked.connect(self.run_time_window_capture)
        hbox.addWidget(b1)
        hbox.addStretch(1)
        main_vbox.addLayout(hbox)

        main_vbox.addStretch(1)
        self.setLayout(main_vbox)

    @QtCore.Slot()
    def run_time_window_capture(self):
        """
        Capture 4D changes over time
        """
        if self.ivm.vol is None:
            error_dialog("No data loaded")
            return

        shape = self.ivm.vol.shape

        # Choose a folder to save images
        fname = QtGui.QFileDialog.getExistingDirectory(self, 'Choose folder to save images')
        if fname == '':
            return

        for ii in range(shape[-1]):
            self.debug("Frame number:", ii)
            self.ivl.set_time_pos(ii)
            output_name = fname + '/' + str(ii).zfill(3) + '.png'
            self.ivl.capture_view_as_image(1, output_name)

QP_WIDGETS = []








"""
Quantiphyse - 2d ortho slice image viewer

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division, unicode_literals, absolute_import

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

import numpy as np
import pyqtgraph as pg

from quantiphyse.utils import LogSource
from quantiphyse.utils.enums import Orientation, DisplayOrder, Visibility, Boundary
from quantiphyse.data import OrthoSlice
from quantiphyse.data.qpdata import Metadata
from quantiphyse.gui.colors import get_lut, get_col

from .maskable_image import MaskableImage

MAIN_DATA = ""

# An arbitrary 'big number' expected to be larger than the largest number of 
# data sets the viewer is ever likely to hold
MAX_NUM_DATA_SETS = 1000

class SliceDataView(LogSource):
    """
    Draws a slice through a data item
    """ 

    def __init__(self, ivm, qpdata, viewbox, plane, vol, view_metadata=None):
        """
        :param qpdata: QpData instance
        :param viewbox: pyqtgraph ViewBox instance
        :param view_metadata: View parameters
        """
        LogSource.__init__(self)
        self._ivm = ivm
        self._qpdata = qpdata
        self._viewbox = viewbox
        self._plane = plane
        self._vol = vol
        self._view = self._qpdata.view
        if view_metadata is not None:
            self._view = view_metadata
        
        self._redraw_options = [
            "visible",
            "roi",
            "z_order",
            "interp_order",
            "contour",
        ]
        self._img = MaskableImage()
        self._contours = []
        self._viewbox.addItem(self._img)
        self._lut = get_lut(self._view.cmap, self._view.alpha)
        self.update()
        self.redraw()
        self._view.sig_changed.connect(self._view_metadata_changed)

    @property
    def plane(self):
        """ Current SlicePlane the viewer is displaying """
        return self._plane

    @plane.setter
    def plane(self, plane):
        if plane != self._plane:
            self.debug("Plane changed - redrawing")
            self._plane = plane
            self.redraw()

    @property
    def vol(self):
        """ Current data volume being displayed """
        return self._vol

    @vol.setter
    def vol(self, vol):
        if vol != self._vol and vol < self._qpdata.nvols:
            self.debug("Volume changed - redrawing")
            self._vol = vol
            self.redraw()

    def _view_metadata_changed(self, key, value):
        self.debug("View params changed: %s=%s", key, value)
        if key in ("cmap", "lut"):
            if self._view.cmap != "custom":
                self._lut = get_lut(self._view.cmap, self._view.alpha)
            else:
                self._lut = self._view.lut
        elif key == "alpha":
            self._lut = [list(row[:3]) + [self._view.alpha] for row in self._lut]
        self.update()
        if key in self._redraw_options:
            self.redraw()

    def update(self):
        """
        Update the image without re-slicing the data
        """
        self.debug("visible? %s", self._view.visible)
        self._img.setVisible(self._view.visible == Visibility.SHOW and (not self._qpdata.roi or bool(self._view.shade)))
        self._img.set_boundary_mode(self._view.boundary)
        self._img.setLookupTable(self._lut, update=True)
        self._img.setLevels(self._view.cmap_range)
        self.debug("overlay cmap range %s", self._view.cmap_range)

    def redraw(self, interp_order=0):
        """
        Update the image when the slice data may have changed
        and therefore may need re-extracting.

        This is more expensive than just changing image parameters so it
        is only triggered when required
        """
        self.debug("slicedataview: redrawing image")
        self.debug(self._vol)
        self.debug("%s, %s", self._plane.basis, self._plane.normal)

        self._z_order = self._view.z_order
        if self._qpdata.roi:
            # FIXME ROIs always on top - should be option
            self._z_order += MAX_NUM_DATA_SETS

        if self._img.isVisible() or self._view.contour:
            slicedata, slicemask, scale, offset = self._qpdata.slice_data(self._plane, vol=self._vol,
                                                                          interp_order=interp_order)
            self.debug("Image data range: %f, %f", np.min(slicedata), np.max(slicedata))
            qtransform = QtGui.QTransform(scale[0, 0], scale[0, 1],
                                          scale[1, 0], scale[1, 1],
                                          offset[0], offset[1])
        if self._img.isVisible():
            self._img.setTransform(qtransform)
            self._img.setImage(slicedata, autoLevels=False)
            self._img.setZValue(self._z_order)

            if self._view.roi:
                roi = self._ivm.data[self._view.roi]
                resampled_roi = roi.resample(self._qpdata.grid)
                maskdata, _, _, _ = resampled_roi.slice_data(self._plane)
                self._img.mask = np.logical_and(maskdata, slicemask)
            else:
                self._img.mask = slicemask

        n_contours = 0
        if self._qpdata.roi and self._view.contour and self._view.visible == Visibility.SHOW:
            # Update data and level for existing contour items, and create new ones if needed
            max_region = max(self._qpdata.regions.keys())
            for val in self._qpdata.regions:
                # Contours do not have alpha transparency
                pencol = get_col(self._lut, val, (1, max_region))[:3]
                if val != 0:
                    if n_contours == len(self._contours):
                        self._contours.append(pg.IsocurveItem())
                        self._viewbox.addItem(self._contours[n_contours])

                    contour = self._contours[n_contours]
                    contour.setTransform(qtransform)
                    contour.setData((slicedata == val).astype(np.int))
                    contour.setLevel(1)
                    contour.setPen(pg.mkPen(pencol, width=3))
                    contour.setZValue(self._z_order)
                    n_contours += 1

        # Clear data from contours not required - FIXME delete them?
        for idx in range(n_contours, len(self._contours)):
            self._contours[idx].setData(None)

    def remove(self):
        """
        Remove the view from the viewbox
        """
        self.debug("Removing slice view")
        self._viewbox.removeItem(self._img)
        for contour in self._contours:
            self._viewbox.removeItem(contour)

class OrthoSliceViewer(pg.GraphicsView, LogSource):
    """
    Displays an orthographic slice through data/ROIs relative to the
    a main viewer grid

    This viewer takes its grid and focus point from a Viewer object
    and maintains its own set of SliceDataView objects. It optionally
    displays the main data as a greyscale background image
    """

    # Signals when point is selected
    sig_pick = QtCore.Signal(int, list)

    # Signals when mouse is draggged. Only emitted if
    # active picker uses drag selection
    sig_drag = QtCore.Signal(int, list)

    # Signals when view is double clicked
    sig_doubleclick = QtCore.Signal(int)

    def __init__(self, ivl, ivm, ax_map, ax_labels):
        """
        :param ivl: Viewer
        :param ivm: ImageVolumeManagement
        :param ax_map: Sequence defining the x, y, z axis of the slice viewer
                       in terms of RAS axis sequence indexes
        :param ax_labels: Sequence of labels for the RAS axes
        """
        LogSource.__init__(self)
        pg.GraphicsView.__init__(self)
        self._ivl = ivl
        self.ivm = ivm
        self.xaxis, self.yaxis, self.zaxis = ax_map
        self._slicez = 0
        self._vol = 0
        self._plane = OrthoSlice(self._ivl.grid, self.zaxis, self._slicez)
        self._main_view = None
        self._dragging = False
        self._arrow_items = []
        self._data_views = {}
        self.debug("axes=%i, %i, %i", self.xaxis, self.yaxis, self.zaxis)

        # View box to display graphics items
        self._viewbox = pg.ViewBox(name="view%i" % self.zaxis, border=pg.mkPen((0x6c, 0x6c, 0x6c), width=2.0))
        self._viewbox.setAspectLocked(True)
        self._viewbox.setBackgroundColor([0, 0, 0])
        self._viewbox.enableAutoRange()
        self.setCentralItem(self._viewbox)

        # Crosshairs
        self._vline = pg.InfiniteLine(angle=90, movable=False)
        self._vline.setZValue(2*MAX_NUM_DATA_SETS)
        self._vline.setPen(pg.mkPen((0, 255, 0), width=1.0, style=QtCore.Qt.DashLine))

        self._hline = pg.InfiniteLine(angle=0, movable=False)
        self._hline.setZValue(2*MAX_NUM_DATA_SETS)
        self._hline.setPen(pg.mkPen((0, 255, 0), width=1.0, style=QtCore.Qt.DashLine))

        self._viewbox.addItem(self._vline, ignoreBounds=True)
        self._viewbox.addItem(self._hline, ignoreBounds=True)

        # Dummy image item which enables us to translate click co-ordinates
        # into image space co-ordinates even when there is no data in the view
        self._dummy = pg.ImageItem()
        self._dummy.setVisible(False)
        self._viewbox.addItem(self._dummy, ignoreBounds=True)

        # Static labels for the view directions
        self._labels = []
        for axis in [self.xaxis, self.yaxis]:
            self._labels.append(QtGui.QLabel(ax_labels[axis][0], parent=self))
            self._labels.append(QtGui.QLabel(ax_labels[axis][1], parent=self))
        for label in self._labels:
            label.setAttribute(QtCore.Qt.WA_TranslucentBackground)

        self._grid_changed(self._ivl.grid)
        self._focus_changed(self._ivl.focus())
        self._update_crosshairs()
        self._update_orientation()

        # Connect to signals from the parent viewer
        self._ivl.sig_grid_changed.connect(self._grid_changed)
        self._ivl.sig_focus_changed.connect(self._focus_changed)
        self._ivl.sig_arrows_changed.connect(self._arrows_changed)
        self._ivl.opts.sig_changed.connect(self._view_opts_changed)

        # Connect to data change signals
        self.ivm.sig_all_data.connect(self._data_changed)
        self.ivm.sig_main_data.connect(self._main_data_changed)

        # Need to intercept the default resize event
        # FIXME why can't call superclass method normally?
        self.resizeEventOrig = self.resizeEvent
        self.resizeEvent = self._window_resized

    def reset(self):
        """ Reset the viewer to show all data sets"""
        self.debug("Auto range")
        self._viewbox.autoRange()

    def redraw(self):
        """ Force a redraw of the viewer, e.g. if data has changed """
        for view in self._data_views.values():
            view.redraw()

    def _view_opts_changed(self, key, value):
        if key in ("orientation", "labels"):
            self._update_orientation()
        elif key == "crosshairs":
            self._update_crosshairs()
        elif key == "main_data":
            self._update_main_data()

    def _grid_changed(self, grid):
        """
        Set the grid that the slice viewer is relative to
        """
        # Adjust axis scaling to that of the viewing grid so voxels have correct relative size
        self._viewbox.setAspectLocked(True, ratio=(grid.spacing[self.xaxis] / grid.spacing[self.yaxis]))
        self._update_slice()
        self.reset()

    def _focus_changed(self, focus):
        self.debug("focus=%s", focus)

        if self.isVisible():
            self._vline.setPos(float(focus[self.xaxis]))
            self._hline.setPos(float(focus[self.yaxis]))

            if focus[self.zaxis] != self._slicez or self._vol != focus[3]:
                self._update_slice()
                self._update_visible_arrows()

    def _arrows_changed(self, arrows):
        """
        Set the locations and colours of arrows to be drawn

        :param arrows: Sequence of tuples: (position, color). Position
                       is a XYZ tuple, color is anything that can
                       be passed to pyqtgraph.mkPen (e.g. color name
                       or RGB tuple)
        """
        item_num = 0
        for pos, col in arrows:
            if item_num == len(self._arrow_items):
                item = pg.ArrowItem()
                self._viewbox.addItem(item)
                self._arrow_items.append((pos, col, item))
            _, _, item = self._arrow_items[item_num]
            item.setPos(float(pos[self.xaxis]), float(pos[self.yaxis]))
            item.setPen(pg.mkPen(col))
            item.setBrush(pg.mkBrush(col))
            item.setZValue(2)
            self._arrow_items[item_num] = (pos, col, item)
            item_num += 1

        for _, _, item in self._arrow_items[item_num:]:
            self._viewbox.removeItem(item)

        self._arrow_items = self._arrow_items[:item_num]
        self._update_visible_arrows()

    def _data_changed(self, data_names):
        self.debug("data changed")
        for name, view in list(self._data_views.items()):
            if name not in data_names and name != MAIN_DATA:
                view.remove()
                del self._data_views[name]

        for name in data_names:
            if name not in self._data_views:
                qpdata = self.ivm.data[name]
                self._data_views[name] = SliceDataView(self.ivm, qpdata, self._viewbox, self._plane, self._vol)

        self._update_crosshairs()
        self._update_orientation()

    def _main_data_changed(self):
        self._update_main_data()

    def _update_main_data(self):
        if MAIN_DATA in self._data_views:
            self._data_views[MAIN_DATA].remove()
            del self._data_views[MAIN_DATA]

        if self.ivm.main is not None and self._ivl.opts.main_data == Visibility.SHOW:
            self._data_views[MAIN_DATA] = SliceDataView(self.ivm, self.ivm.main, self._viewbox, self._plane, self._vol, self._ivl.main_view_md)
        self.reset()

    def _update_crosshairs(self):
        crosshairs_visible = len(self._data_views) > 0 and self._ivl.opts.crosshairs == Visibility.SHOW
        self._vline.setVisible(crosshairs_visible)
        self._hline.setVisible(crosshairs_visible)

    def _update_orientation(self):
        labels_visible = len(self._data_views) > 0 and self._ivl.opts.labels == Visibility.SHOW
        for label in self._labels:
            label.setVisible(labels_visible)

        if self.xaxis == 0:
            if self._ivl.opts.orientation == Orientation.NEUROLOGICAL:
                left, right, invert = 0, 1, False
            else:
                left, right, invert = 1, 0, True
            self._viewbox.invertX(invert)
            self._labels[right].setText("R")
            self._labels[left].setText("L")

    def _update_slice(self):
        self._slicez = self._ivl.focus()[self.zaxis]
        self._vol = self._ivl.focus()[3]
        self._plane = OrthoSlice(self._ivl.grid, self.zaxis, self._slicez)
        for view in self._data_views.values():
            view.plane = self._plane
            view.vol = self._vol
        self.debug("set slice: %f %i", self._slicez, self._vol)

    def _update_visible_arrows(self):
        """
        Update arrows so only those visible are shown
        """
        current_zpos = int(self._ivl.focus()[self.zaxis] + 0.5)
        for pos, _, item in self._arrow_items:
            arrow_zpos = int(pos[self.zaxis] + 0.5)
            item.setVisible(current_zpos == arrow_zpos)

    def _window_resized(self, event):
        """
        Called when window is resized - updates the position
        of the text labels and then calls the original resize method
        """
        w = self.geometry().width()
        h = self.geometry().height()
        self._labels[0].setGeometry(0, h/2, 10, 10)
        self._labels[1].setGeometry(w-10, h/2, 10, 10)
        self._labels[2].setGeometry(w/2, h-10, 10, 10)
        self._labels[3].setGeometry(w/2, 0, 10, 10)
        self.resizeEventOrig(event)

    def wheelEvent(self, event):
        """
        Subclassed to remove scroll to zoom from pg.ImageItem
        and instead trigger a scroll through the volume
        """
        dz = int(event.delta()/120)
        pos = self._ivl.focus()
        pos[self.zaxis] += dz
        self._ivl.set_focus(pos)

    def mousePressEvent(self, event):
        """
        Called when mouse button is pressed on the view
        """
        super(OrthoSliceViewer, self).mousePressEvent(event)
        if self.ivm.main is None:
            return

        if event.button() == QtCore.Qt.LeftButton:
            # Convert co-ords to view grid
            coords = self._dummy.mapFromScene(event.pos())
            pos = self._ivl.focus()
            pos[self.xaxis] = coords.x()
            pos[self.yaxis] = coords.y()
            self._ivl.set_focus(pos)

            if self._ivl.picker.use_drag:
                self._dragging = True
            self.sig_pick.emit(self.zaxis, self._ivl.focus())

    def mouseReleaseEvent(self, event):
        """
        Called when mouse button is released on the view
        """
        super(OrthoSliceViewer, self).mouseReleaseEvent(event)
        self._dragging = False

    def mouseDoubleClickEvent(self, event):
        """
        Called when mouse button is double clicked on the view

        This is used to maximise/minimise the view window
        """
        super(OrthoSliceViewer, self).mouseDoubleClickEvent(event)
        if event.button() == QtCore.Qt.LeftButton:
            self.sig_doubleclick.emit(self.zaxis)

    def mouseMoveEvent(self, event):
        """
        Called when the mouse is moved over the window.

        The default behaviour is used unless we are dragging a
        selection region
        """
        if self._dragging:
            coords = self._dummy.mapFromScene(event.pos())
            pos = self._ivl.focus()
            pos[self.xaxis] = coords.x()
            pos[self.yaxis] = coords.y()
            self.sig_drag.emit(self.zaxis, pos)
        else:
            super(OrthoSliceViewer, self).mouseMoveEvent(event)



"""
Quantiphyse - tests for ImageVolumeManagement

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest
import numpy as np

from quantiphyse.data import ImageVolumeManagement, NumpyData, DataGrid

GRIDSIZE = 5

class IVMTest(unittest.TestCase):

    def setUp(self):
        self.ivm = ImageVolumeManagement()

    def testEmptyCreate(self):
        self.assertEqual(len(self.ivm.data), 0)
        self.assertEqual(len(self.ivm.rois), 0)
        self.assertTrue(self.ivm.current_data is None)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertTrue(self.ivm.main is None)
        
    def testAdd(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        self.ivm.add(qpd)
        self.assertEqual(len(self.ivm.data), 1)
        self.assertEqual(len(self.ivm.rois), 0)
        self.assertTrue(self.ivm.current_data is None)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.data["test"], qpd)

    def testAddRoi(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.randint(0, 10, size=shape), name="test", grid=grid, roi=True)
        self.assertTrue(qpd.roi)
        self.ivm.add(qpd)
        self.assertEqual(len(self.ivm.data), 1)
        self.assertEqual(len(self.ivm.rois), 1)
        self.assertTrue(self.ivm.current_data is None)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.data["test"], qpd)

    def testAddTwo(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        qpd2 = NumpyData(np.random.rand(*shape), name="test2", grid=grid)
        self.ivm.add(qpd)
        self.ivm.add(qpd2)
        self.assertEqual(len(self.ivm.data), 2)
        self.assertEqual(len(self.ivm.rois), 0)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.current_data, qpd2)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertEqual(self.ivm.data["test"], qpd)
        self.assertEqual(self.ivm.data["test2"], qpd2)

    def testAddTwoRois(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.randint(0, 10, size=shape), name="test", grid=grid, roi=True)
        qpd2 = NumpyData(np.random.randint(0, 10, size=shape), name="test2", grid=grid, roi=True)
        self.assertTrue(qpd.roi)
        self.assertTrue(qpd2.roi)
        self.ivm.add(qpd)
        self.ivm.add(qpd2)
        self.assertEqual(len(self.ivm.data), 2)
        self.assertEqual(len(self.ivm.rois), 2)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.current_roi, qpd2)
        self.assertTrue(self.ivm.current_data is None)
        self.assertEqual(self.ivm.data["test"], qpd)
        self.assertEqual(self.ivm.data["test2"], qpd2)
        self.assertEqual(self.ivm.rois["test"], qpd)
        self.assertEqual(self.ivm.rois["test2"], qpd2)

    def testAddTwoMixed1(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        qpd2 = NumpyData(np.random.randint(0, 10, size=shape), name="test2", grid=grid, roi=True)
        self.assertFalse(qpd.roi)
        self.assertTrue(qpd2.roi)
        self.ivm.add(qpd)
        self.ivm.add(qpd2)
        self.assertEqual(len(self.ivm.data), 2)
        self.assertEqual(len(self.ivm.rois), 1)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.current_roi, qpd2)
        self.assertTrue(self.ivm.current_data is None)
        self.assertEqual(self.ivm.data["test"], qpd)
        self.assertEqual(self.ivm.data["test2"], qpd2)
        self.assertEqual(self.ivm.rois["test2"], qpd2)

    def testAddTwoMixed2(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        qpd2 = NumpyData(np.random.randint(0, 10, size=shape), name="test2", grid=grid, roi=True)
        self.assertFalse(qpd.roi)
        self.assertTrue(qpd2.roi)
        self.ivm.add(qpd2)
        self.ivm.add(qpd)
        self.assertEqual(len(self.ivm.data), 2)
        self.assertEqual(len(self.ivm.rois), 1)
        self.assertEqual(self.ivm.main, qpd2)
        self.assertEqual(self.ivm.current_data, qpd)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertEqual(self.ivm.data["test"], qpd)
        self.assertEqual(self.ivm.data["test2"], qpd2)
        self.assertEqual(self.ivm.rois["test2"], qpd2)

    def testAddThreeMixed(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        qpd2 = NumpyData(np.random.rand(*shape), name="test2", grid=grid)
        qpd3 = NumpyData(np.random.randint(0, 10, size=shape), name="test3", grid=grid, roi=True)
        self.assertFalse(qpd.roi)
        self.assertFalse(qpd2.roi)
        self.assertTrue(qpd3.roi)
        self.ivm.add(qpd)
        self.ivm.add(qpd2)
        self.ivm.add(qpd3)
        self.assertEqual(len(self.ivm.data), 3)
        self.assertEqual(len(self.ivm.rois), 1)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.current_data, qpd2)
        self.assertEqual(self.ivm.current_roi, qpd3)
        self.assertEqual(self.ivm.data["test"], qpd)
        self.assertEqual(self.ivm.data["test2"], qpd2)
        self.assertEqual(self.ivm.data["test3"], qpd3)
        self.assertEqual(self.ivm.rois["test3"], qpd3)

    def testDelete(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        self.ivm.add(qpd)
        self.assertEqual(len(self.ivm.data), 1)
        self.assertEqual(len(self.ivm.rois), 0)
        self.assertTrue(self.ivm.current_data is None)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertEqual(self.ivm.main, qpd)
        self.assertEqual(self.ivm.data["test"], qpd)
        self.ivm.delete(qpd.name)
        self.assertEqual(len(self.ivm.data), 0)
        self.assertEqual(len(self.ivm.rois), 0)
        self.assertTrue(self.ivm.current_data is None)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertTrue(self.ivm.main is None)
      
    def testRename(self):
        shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        grid = DataGrid(shape, np.identity(4))
        qpd = NumpyData(np.random.rand(*shape), name="test", grid=grid)
        self.ivm.add(qpd)
        self.ivm.rename("test", "test2")
        self.assertEqual(len(self.ivm.data), 1)
        self.assertEqual(len(self.ivm.rois), 0)
        self.assertTrue(self.ivm.current_data is None)
        self.assertTrue(self.ivm.current_roi is None)
        self.assertEqual(self.ivm.main, self.ivm.data["test2"])
        self.assertTrue(np.all(self.ivm.data["test2"].raw() == qpd.raw()))

if __name__ == '__main__':
    unittest.main()



"""
Quantiphyse - tests for QpData classes

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import unittest
import tempfile

import numpy as np

from quantiphyse.data import NumpyData, DataGrid
import quantiphyse.data.nifti as nifti

GRIDSIZE = 5
NVOLS = 4

class NumpyDataTest(unittest.TestCase):
    """ Tests for the NumpyData subclass of QpData """

    def setUp(self):
        self.shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        self.grid = DataGrid(self.shape, np.identity(4))
        self.floats = np.random.rand(*self.shape)
        self.ints = np.random.randint(0, 10, self.shape)
        self.floats4d = np.random.rand(*(self.shape + [NVOLS,]))

    def test3d(self):
        qpd = NumpyData(self.floats, grid=self.grid, name="test")
        self.assertEqual(qpd.name, "test")
        self.assertEqual(qpd.nvols, 1)
        self.assertEqual(qpd.ndim, 3)
        self.assertFalse(qpd.roi)
        
    def test3dints(self):
        qpd = NumpyData(self.ints, grid=self.grid, name="test", roi=True)
        self.assertEqual(qpd.name, "test")
        self.assertEqual(qpd.nvols, 1)
        self.assertEqual(qpd.ndim, 3)
        self.assertTrue(qpd.roi)
        
    def test4d(self):
        qpd = NumpyData(self.floats4d, grid=self.grid, name="test")
        self.assertEqual(qpd.name, "test")
        self.assertEqual(qpd.nvols, NVOLS)
        self.assertEqual(qpd.ndim, 4)
        self.assertFalse(qpd.roi)
        
    def testRandomName(self):
        name = "sdkfjhsdfi4"
        qpd = NumpyData(self.floats4d, grid=self.grid, name=name)
        self.assertEqual(qpd.name, name)
        
    def testRoiFloats(self):
        """ Check that ROIs can contain float data so long as the numbers are really integers """
        qpd = NumpyData(self.ints.astype(np.float), grid=self.grid, name="test", roi=True)
        self.assertTrue(issubclass(qpd.raw().dtype.type, np.floating))
        self.assertTrue(qpd.roi)
        
    def testRoiRegions(self):
        qpd = NumpyData(self.ints, grid=self.grid, name="test", roi=True)
        self.assertTrue(qpd.roi)
        regions = [v for v in np.unique(self.ints) if v != 0]
        self.assertEqual(regions, list(qpd.regions.keys()))

    def testRange(self):
        qpd = NumpyData(self.floats, grid=self.grid, name="test")
        mx, mn = np.max(self.floats), np.min(self.floats)
        self.assertAlmostEqual(qpd.range()[0], mn)
        self.assertAlmostEqual(qpd.range()[1], mx)
        
    def testSet2dt(self):
        qpd = NumpyData(self.floats, grid=self.grid, name="test")
        qpd.set_2dt()
        self.assertEqual(qpd.nvols, GRIDSIZE)
        self.assertEqual(qpd.ndim, 4)
        d = qpd.raw()
        for idx in range(GRIDSIZE):
            vol = d[..., idx]
            self.assertEqual(vol.shape[0], GRIDSIZE)
            self.assertEqual(vol.shape[1],  GRIDSIZE)
            self.assertEqual(vol.shape[2], 1)
            self.assertTrue(np.allclose(np.squeeze(vol), self.floats[:, :, idx]))

    def testVolume(self):
        qpd = NumpyData(self.floats4d, grid=self.grid, name="test")
        for idx in range(NVOLS):
            self.assertTrue(np.allclose(qpd.volume(idx), self.floats4d[..., idx]))
        
    def testValue3d(self):
        qpd = NumpyData(self.floats, grid=self.grid, name="test")
        POS = [2, 3, 4]
        self.assertAlmostEqual(qpd.value(POS), self.floats[POS[0], POS[1], POS[2]])
           
    def testValue3d4dPos(self):
        qpd = NumpyData(self.floats, grid=self.grid, name="test")
        POS = [2, 3, 4, 1]
        self.assertAlmostEqual(qpd.value(POS), self.floats[POS[0], POS[1], POS[2]])
     
    def testValue4d(self):
        qpd = NumpyData(self.floats4d, grid=self.grid, name="test")
        POS = [2, 3, 4, 1]
        self.assertAlmostEqual(qpd.value(POS), self.floats4d[POS[0], POS[1], POS[2], POS[3]])
        
    def testValue4d3dPos(self):
        qpd = NumpyData(self.floats4d, grid=self.grid, name="test")
        POS = [2, 3, 4]
        self.assertAlmostEqual(qpd.value(POS), self.floats4d[POS[0], POS[1], POS[2], 0])
        
class NiftiDataTest(unittest.TestCase):
    """ Tests for the NiftiData subclass of QpData """

    def setUp(self):
        self.shape = [GRIDSIZE, GRIDSIZE, GRIDSIZE]
        self.grid = DataGrid(self.shape, np.identity(4))
        self.floats = np.random.rand(*self.shape)
        self.ints = np.random.randint(0, 10, self.shape)
        self.floats4d = np.random.rand(*(self.shape + [NVOLS,]))

    def testSaveLoadMetadata(self):
        tempdir = tempfile.mkdtemp(prefix="qp")
        fname = os.path.join(tempdir, "test.nii.gz")
        md = {
            "flibble" : 2.3,
            "flobble" : 4,
            "fishcakes" : ["a", "b", "c"],
            "fluntycaps" : {
                "gfruffle" : "def",
                "gfable" : 8,
            }
        }
        qpd = NumpyData(self.floats4d, grid=self.grid, name="test", metadata=md)
        nifti.save(qpd, fname)

        nifti_data = nifti.NiftiData(fname)
        for key, value in md.items():
            self.assertTrue(key in nifti_data.metadata)
            self.assertEqual(nifti_data.metadata[key], value)

    def testLoadSaveSameName(self):
        qpd = NumpyData(self.floats4d, grid=self.grid, name="test")

        tempdir = tempfile.mkdtemp(prefix="qp")
        fname = os.path.join(tempdir, "test.nii")
        nifti.save(qpd, fname)

        nifti_data = nifti.NiftiData(fname)
        nifti.save(nifti_data, fname)

if __name__ == '__main__':
    unittest.main()



"""
Quantiphyse - Data management framework

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division

import logging
import keyword
import re
from collections import OrderedDict

try:
    from PySide import QtCore
except ImportError:
    from PySide2 import QtCore

import numpy as np

from quantiphyse.utils import QpException
from quantiphyse.utils.enums import Visibility

from .qpdata import QpData
from .load_save import NumpyData
from .extras import Extra

LOG = logging.getLogger(__name__)

class ImageVolumeManagement(QtCore.QObject):
    """
    Holds all image datas used in analysis

    Has to inherit from a Qt base class that supports signals

    Attributes
    ----------

      ``main`` 'main' QpData - typically viewed as a greyscale background
      ``data`` Mapping from data set name to QpData object
      ``current_data`` QpData which is the 'current' overlay data set
      ``current_roi`` QpData with ``roi=True`` used as the current ROI
      ``extras`` Mapping from name to object for miscellaneous extra data.
                 Extras must support string-conversion for writing to files.
    """
    # Signals

    # Change to main data
    sig_main_data = QtCore.Signal(object)

    # Change to current data
    sig_current_data = QtCore.Signal(object)

    # Change to set of data (e.g. new one added)
    sig_all_data = QtCore.Signal(list)

    # Change to current ROI
    sig_current_roi = QtCore.Signal(object)

    # Change to set of extras (e.g. new one added)
    sig_extras = QtCore.Signal(list)

    def __init__(self):
        super(ImageVolumeManagement, self).__init__()
        self.reset()

    def reset(self):
        """ Clear all data """
        self.main = None
        self.data = OrderedDict()
        self.current_data = None
        self.current_roi = None
        self.extras = OrderedDict()

        self.sig_main_data.emit(None)
        self.sig_current_data.emit(None)
        self.sig_current_roi.emit(None)
        self.sig_all_data.emit([])

    @property
    def rois(self):
        """
        :return: Dictionary of name : QpData for all data items which can be used as ROIs
        """
        return dict([(data.name, data) for data in self.data.values() if data.roi])

    def suggest_name(self, name, ensure_unique=True):
        """
        Suggest a name for new data that is suitable for use as a Python variable.
        If required, ensure name does not clash with existing names
        """
        # Remove invalid characters and make sure there's something left
        name = re.sub('[^0-9a-zA-Z_]', '', name)
        if not name:
            name = "data"

        # Add underscore if it's a keyword or starts with a digit
        if name[0] in ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9') or keyword.iskeyword(name):
            name = "_" + name

        # Make it unique
        num = 1
        test_name = name
        while 1:
            if not ensure_unique or (test_name not in self.data):
                break
            num += 1
            test_name = "%s_%i" % (name, num)
        return test_name

    def _valid_name(self, name):
        if name is None or not re.match(r'[a-z_]\w*$', name, re.I) or keyword.iskeyword(name):
            raise QpException("'%s' is not a valid name" % name)

    def set_main_data(self, name):
        """
        Set the named data item as the main data

        :param name: Name of main data. Must be name of data item in IVM
        """
        self._data_exists(name)
        self.main = self.data[name]
        self.sig_main_data.emit(self.main)

    def add(self, data, name=None, grid=None, make_current=None, make_main=None, roi=None):
        """
        Add data item to IVM

        Data will be made the main data if make_main is True, or there is no main data

        Data will be made current if make_current is not specified and there is no
        current data, and make_main is not True

        :param data: Numpy array of QpData instance
        :param name: Name which must be a valid name not already used in this IVM
        :param grid: If data is a Numpy array, a DataGrid instance defining the orientation
        :param make_current: If True, make this the current data item
        :param make_main: If True, make this the main data.
        :param roi: If providing Numpy array, optionally specifies whether the data is an ROI or not
        """
        if isinstance(data, np.ndarray):
            if grid is None or name is None:
                raise RuntimeError("add: Numpy data must have a name and a grid")
            data = NumpyData(data, grid, name, roi=roi)
        elif not isinstance(data, QpData):
            raise QpException("add: data must be Numpy array or QpData")

        if name is not None:
            data.name = name

        self._valid_name(data.name)

        # If replacing existing data, delete the old one first
        if data.name in self.data:
            if self.current_data is not None and self.current_data.name == data.name:
                self.set_current_data(None)
                make_current = True
            if self.current_roi is not None and self.current_roi.name == data.name:
                self.set_current_roi(None)
                make_current = True
            if self.main is not None and self.main.name == data.name:
                make_main = True
            del self.data[data.name]
            self.sig_all_data.emit(list(self.data.keys()))

        self.data[data.name] = data

        # Set z-order
        data.view.z_order = len(self.data)

        # Emit the 'data changed' signal
        self.sig_all_data.emit(list(self.data.keys()))

        # Make main data if requested or if not specified and there is no current main data
        if make_main is None:
            make_main = self.main is None
        if make_main:
            self.set_main_data(data.name)
            data.view.visible = Visibility.HIDE
            
        # Make current if requested, or if not specified and it is the first non-main data/ROI
        if make_current is None:
            make_current = ((data.roi and self.current_roi is None) or
                            (not data.roi and self.current_data is None)) and not make_main
        if make_current:
            if data.roi:
                self.set_current_roi(data.name)
            else:
                self.set_current_data(data.name)

    def _data_exists(self, name):
        if name not in self.data:
            raise RuntimeError("Data '%s' does not exist" % name)

    def _roi_exists(self, name):
        if name not in self.rois:
            raise RuntimeError("ROI '%s' does not exist" % name)

    def is_main_data(self, qpd):
        """
        :param qpd: QpData instance
        :return: True if qpd is the main data
        """
        return self.main is not None and qpd is not None and self.main.name == qpd.name

    def is_current_data(self, qpd):
        """
        :param qpd: QpData instance
        :return: True if qpd is the current data
        """
        return self.current_data is not None and qpd is not None and self.current_data.name == qpd.name

    def is_current_roi(self, roi):
        """
        :param qpd: QpData instance
        :return: True if qpd is the current ROI
        """
        return self.current_roi is not None and roi is not None and self.current_roi.name == roi.name

    def set_current_data(self, name):
        """
        Set a named data item as the current data

        :param name: Name of data item which must exist within the IVM
        """
        if name is not None:
            self._data_exists(name)
            self.current_data = self.data[name]
        else:
            self.current_data = None
        self.sig_current_data.emit(self.current_data)

    def set_current(self, name):
        """
        Set a named data item as current (ROI or data as required)
        """
        if name is not None:
            self._data_exists(name)
            if self.data[name].roi:
                self.set_current_roi(name)
            else:
                self.set_current_data(name)
        else:
            self.set_current_data(None)
            self.set_current_roi(None)

    def rename(self, name, newname):
        """
        Rename a data item

        :param name: Name of data item which must exist within the IVM
        :param newname: New name for this data item
        """
        self._data_exists(name)
        qpd = self.data[name]
        qpd.name = newname
        del self.data[name]
        self.data[newname] = qpd
        self.sig_all_data.emit(list(self.data.keys()))

    def delete(self, name):
        """
        Delete a data item

        :param name: Name of data item which must exist within the IVM
        """
        self._data_exists(name)
        del self.data[name]
        if self.current_data is not None and self.current_data.name == name:
            self.current_data = None
            self.sig_current_data.emit(None)
        if self.main is not None and self.main.name == name:
            self.main = None
            self.sig_main_data.emit(None)
        self.sig_all_data.emit(list(self.data.keys()))

    def set_current_roi(self, name):
        """
        Set a named ROI as the current ROI

        :param name: Name of ROI which must exist within the IVM
        """
        if name is not None:
            self._roi_exists(name)
            self.current_roi = self.rois[name]
        else:
            self.current_roi = None
        self.sig_current_roi.emit(self.current_roi)

    def add_extra(self, name, obj):
        """
        Add an 'extra', which can be any result of a process which
        is not voxel data, e.g. a number, table, etc.

        Extras are required to support str() conversion so they
        can be written to a file

        :param name: Name to give the extra. If an extra already exists with this name
                     it will be overwritten
        :param obj: instance of Extra which should support str() conversion
        """
        if not isinstance(obj, Extra):
            raise ValueError("extra object must be subclasses of Extra")
        self.extras[name] = obj
        self.sig_extras.emit(self.extras.values())

    def search_extras(self, cls=None):
        """
        Search extras for those that match given criteria
        
        :param cls: Return extras that are instances of this class
        """
        matching = []
        for extra in self.extras.values():
            if cls is not None and not isinstance(extra, cls):
                continue
            matching.append(extra)
        return matching
        
    def values(self, pos, grid=None):
        """
        Get all the 3D data values at the current position

        :param pos: Position as a 3D or 4D vector. If 4D last value is the volume index
                    (0 for 3D). If ``grid`` not specified, position is in world space
        :param grid: If specified, interpret position in this ``DataGrid`` co-ordinate space.
        :return: Dictionary of data name : value
        """
        values = {}

        # loop over all loaded data and save values in a dictionary
        for name, qpd in self.data.items():
            if qpd.nvols == 1:
                values[name] = qpd.value(pos, grid)

        return values

    def timeseries(self, pos, grid=None):
        """
        Return time/volume series curves for all 4D data items

        :param pos: Position as a 3D or 4D vector. If 4D last value is the volume index
                    (0 for 3D). If ``grid`` not specified, position is in world space
        :param grid: If specified, interpret position in this ``DataGrid`` co-ordinate space.
        :return: Dictionary of data name : sequence of values
        """
        timeseries = {}
        for qpd in self.data.values():
            if qpd.nvols > 1:
                timeseries[qpd.name] = qpd.timeseries(pos, grid)

        return timeseries



"""
Quantiphyse - Viewer for 3D and 4D data

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division, unicode_literals, absolute_import

try:
    from PySide import QtGui, QtCore
except ImportError:
    from PySide2 import QtGui, QtCore

import numpy as np

from quantiphyse.utils import LogSource
from quantiphyse.data.qpdata import DataGrid, Metadata
from quantiphyse.utils.enums import Orientation, DisplayOrder, Visibility, Boundary

from .pickers import PICKERS, PointPicker
from .slice_viewer import OrthoSliceViewer
from .histogram_widget import HistogramWidget, CurrentDataHistogramWidget
from .view_params_widget import ViewParamsWidget
from .navigators import NavigationBox

DEFAULT_MAIN_VIEW = {
    "visible" : Visibility.SHOW,
    "roi_only" : False,
    "boundary" : Boundary.CLAMP,
    "alpha" : 255,
    "cmap" : "grey",
    "z_order" : 0,
}

class SingleViewEnforcer:
    """
    Enforces single-overlay view mode

    This is a kind of kludge class to make the viewer behave like it used to
    i.e. with only one overlay/roi visible at a time. It does this by continuously
    monitoring the data in the IVM and making sure the 'current' data/roi is the only
    one visible.
    """
    def __init__(self, ivm):
        self._ivm = ivm
        self._data_changed(self._ivm.data.keys())
        self._ivm.sig_all_data.connect(self._data_changed)
        self._ivm.sig_current_data.connect(self._current_changed)
        self._ivm.sig_current_roi.connect(self._current_changed)

    def __del__(self):
        self.die()

    def die(self):
        """
        Stop enforcing single view mode. Normally called prior to deletion
        but allow it to be called explicitly so we don't rely on the timing of __del__
        """
        if self._ivm is not None:
            self._ivm.sig_all_data.disconnect(self._data_changed)
            self._ivm.sig_current_data.disconnect(self._current_changed)
            self._ivm.sig_current_roi.disconnect(self._current_changed)
            self._ivm = None

    def _current_changed(self, qpdata):
        if qpdata is not None:
            qpdata.view.visible = Visibility.SHOW
            self._hide_others(qpdata)

    def _data_changed(self, data_names):
        for name in data_names:
            qpdata = self._ivm.data[name]
            if qpdata.view.visible == Visibility.SHOW and \
                qpdata != self._ivm.current_data and \
                qpdata != self._ivm.current_roi:
                qpdata.view.visible = Visibility.HIDE

    def _hide_others(self, qpdata):
        for data in self._ivm.data.values():
            if data.name != qpdata.name and data.roi == qpdata.roi and data.view.visible == Visibility.SHOW:
                data.view.visible = Visibility.HIDE

class Viewer(QtGui.QSplitter, LogSource):
    """
    Widget containing three orthogonal slice views, two histogram/LUT widgets plus
    navigation sliders and data summary view.

    The viewer maintains two main pieces of data: a grid defining the main co-ordinate
    system of the viewer and a point of focus, in co-ordinates relative to the viewing grid.

    In addition, the viewer supports 'arrows' to mark positions in space, and variable
    pickers which control the selection of data.

    The grid is generally either a straightforward 1mm RAS grid, or an approximate RAS grid
    derived from the grid of the main data. Although the focus position is provided and set
    according to this grid by default, the ``focus`` and ``set_focus`` methods allow for
    the co-ordinates to be set or retrieved according to another arbitrary grid.

    :ivar grid: Grid the Viewer uses as the basis for the orthogonal slices.
                This is typically an RAS-aligned version of the main data grid, or
                alternatively an RAS world-grid
    """

    # Signal emitted when point of focus is changed
    sig_focus_changed = QtCore.Signal(list)

    # Signal emitted when point of focus is changed
    sig_grid_changed = QtCore.Signal(object)

    # Signal emitted when the set of marker arrows has changed
    sig_arrows_changed = QtCore.Signal(list)

    # Signal emitted when the picker mode is changed
    sig_picker_changed = QtCore.Signal(object)

    # Signale mitted when the set of selected points is changed.
    # Emission of this signal depends on the picking mode selected
    sig_selection_changed = QtCore.Signal(object)

    def __init__(self, ivm):
        LogSource.__init__(self)
        QtGui.QSplitter.__init__(self, QtCore.Qt.Vertical)

        self.ivm = ivm
        self._grid = DataGrid([1, 1, 1], np.identity(4))
        self._focus = [0, 0, 0, 0]
        self._arrows = []
        self._picker = PointPicker(self)
        self._single_view_enforcer = None
        self.opts = Metadata()
        self.opts.orientation = Orientation.RADIOLOGICAL
        self.opts.display_order = DisplayOrder.USER
        self.opts.crosshairs = Visibility.SHOW
        self.opts.labels = Visibility.SHOW
        self.opts.main_data = Visibility.SHOW
        self.opts.interp = 0

        # Create three orthogonal slice viewers
        # For each viewer, we pass the xyz axis mappings and the labels
        ax_map = [[0, 1, 2], [0, 2, 1], [1, 2, 0]]
        ax_labels = [("L", "R"), ("P", "A"), ("I", "S")]
        self.ortho_views = {}
        for i in range(3):
            win = OrthoSliceViewer(self, self.ivm, ax_map[i], ax_labels)
            win.sig_pick.connect(self._pick)
            win.sig_drag.connect(self._drag)
            win.sig_doubleclick.connect(self._toggle_maximise)
            self.ortho_views[win.zaxis] = win

        # Histogram which controls colour map and levels for main volume
        self.main_histogram = HistogramWidget(self)
        self.main_view_md = Metadata(DEFAULT_MAIN_VIEW)

        # Histogram which controls colour map and levels for the current data
        self.data_histogram = CurrentDataHistogramWidget(self)

        # Layout of the ortho slice viewers and histograms
        gview = QtGui.QWidget()
        self.layout_grid = QtGui.QGridLayout()
        gview.setLayout(self.layout_grid)

        self.layout_grid.setHorizontalSpacing(2)
        self.layout_grid.setVerticalSpacing(2)
        self.layout_grid.setContentsMargins(0, 0, 0, 0)
        self.layout_grid.addWidget(self.ortho_views[1], 0, 0,)
        self.layout_grid.addWidget(self.ortho_views[0], 0, 1)
        self.layout_grid.addWidget(self.ortho_views[2], 1, 0)
        self.layout_grid.addWidget(self.main_histogram, 0, 2)
        self.layout_grid.addWidget(self.data_histogram, 1, 2)
        self.layout_grid.setColumnStretch(0, 3)
        self.layout_grid.setColumnStretch(1, 3)
        self.layout_grid.setColumnStretch(2, 1)
        self.layout_grid.setRowStretch(0, 1)
        self.layout_grid.setRowStretch(1, 1)

        # Navigation controls layout
        control_box = QtGui.QWidget()
        vbox = QtGui.QVBoxLayout()
        vbox.setSpacing(5)
        control_box.setLayout(vbox)

        # Navigation sliders and the ROI/Overlay view controls
        #vbox.addWidget(DataSummary(self))
        hbox = QtGui.QHBoxLayout()
        nav_box = NavigationBox(self)
        hbox.addWidget(nav_box)
        roi_box = ViewParamsWidget(self, rois=True, data=False)
        hbox.addWidget(roi_box)
        ovl_box = ViewParamsWidget(self, rois=False, data=True)
        hbox.addWidget(ovl_box)
        vbox.addLayout(hbox)

        # Overall layout with the viewers above and the controls below
        self.addWidget(gview)
        self.addWidget(control_box)
        self.setStretchFactor(0, 5)
        self.setStretchFactor(1, 1)

        # Connect to signals
        self.ivm.sig_main_data.connect(self._main_data_changed)

    def redraw(self):
        """
        Force redraw of all data
        """
        for view in self.ortho_views.values():
            view.redraw()

    @property
    def picker(self):
        """ Current picker object """
        return self._picker

    @property
    def grid(self):
        """ DataGrid object used as the fundamental grid for the viewer """
        return self._grid

    @property
    def arrows(self):
        """ Sequence of arrows defined by the viewer """
        return self._arrows

    @property
    def show_main(self):
        """ True if main grid data is being displayed as a greyscale background """
        return self.main_view_md.visible == Visibility.SHOW

    @show_main.setter
    def show_main(self, show_main):
        self._main_view_md.visible = show_main

    @property
    def multiview(self):
        return self._single_view_enforcer is None

    @multiview.setter
    def multiview(self, mv):
        if self.multiview == bool(mv):
            return
        elif mv:
            self._single_view_enforcer.die()
            self._single_view_enforcer = None
        else:
            self._single_view_enforcer = SingleViewEnforcer(self.ivm)

    def focus(self, grid=None):
        """
        Get the current focus position

        :param grid: Report position using co-ordinates relative to this grid.
                     If not specified, report current view grid co-ordinates
        :return: 4D sequence containing position plus the current data volume index
        """
        if grid is None:
            return list(self._focus)
        else:
            world = self._grid.grid_to_world(self._focus)
            return list(grid.world_to_grid(world))

    def set_focus(self, pos, grid=None):
        """
        Set the current focus position

        :param grid: Specify position using co-ordinates relative to this grid.
                     If not specified, position is in current view grid co-ordinates
        """
        if grid is not None:
            world = grid.grid_to_world(pos)
            pos = self._grid.world_to_grid(world)

        self._focus = list(pos)
        if len(self._focus) != 4:
            raise Exception("Position must be 4D")

        self.debug("Cursor position: %s", self._focus)
        self.sig_focus_changed.emit(self._focus)

    def set_picker(self, pickmode):
        """
        Deprecated
        """
        self.set_pickmode(pickmode)

    def set_pickmode(self, pickmode):
        """
        Set the picking mode

        :param pickmode: Picking mode from :class:`PickMode`
        """
        self._picker.cleanup()
        self._picker = PICKERS[pickmode](self)
        self.sig_picker_changed.emit(self._picker)

    def add_arrow(self, pos, grid=None, col=None):
        """
        Add an arrow to mark a particular position

        :param pos:  Position co-ordinates
        :param grid: Grid co-ordinates are relative to, if not specified
                     uses viewing grid
        :param col:  Colour as RGB sequence, if not specified uses a default
        """
        if grid is not None:
            world = grid.grid_to_world(pos)
            pos = self._grid.world_to_grid(world)

        if col is None:
            # Default to grey arrow
            col = [127, 127, 127]

        self._arrows.append((pos, col))
        self.sig_arrows_changed.emit(self._arrows)

    def remove_arrows(self):
        """
        Remove all the arrows that have been placed
        """
        self._arrows = []
        self.sig_arrows_changed.emit(self._arrows)

    def _pick(self, win, pos):
        """
        Called when a point is picked in one of the viewing windows
        """
        self._picker.pick(win, pos)
        self.sig_selection_changed.emit(self._picker)

    def _drag(self, win, pos):
        """
        Called when a drag selection is changed in one of the viewing windows
        """
        self._picker.drag(win, pos)
        self.sig_selection_changed.emit(self._picker)

    def _toggle_maximise(self, win, state=-1):
        """
        Maximise/Minimise view window
        If state=1, maximise, 0=show all, -1=toggle
        """
        win1 = (win+1) % 3
        win2 = (win+2) % 3
        if state == 1 or (state == -1 and self.ortho_views[win1].isVisible()):
            # Maximise
            self.layout_grid.addWidget(self.ortho_views[win], 0, 0, 2, 2)
            self.ortho_views[win1].setVisible(False)
            self.ortho_views[win2].setVisible(False)
            self.ortho_views[win].setVisible(True)
        elif state == 0 or (state == -1 and not self.ortho_views[win1].isVisible()):
            # Show all three
            self.layout_grid.addWidget(self.ortho_views[1], 0, 0)
            self.layout_grid.addWidget(self.ortho_views[0], 0, 1)
            self.layout_grid.addWidget(self.ortho_views[2], 1, 0)
            for oview in range(3):
                self.ortho_views[oview].setVisible(True)
                self.ortho_views[oview].update()

    def _main_data_changed(self, data):
        if data is not None:
            self._grid = data.grid.get_standard()
            self.debug("Main data raw grid")
            self.debug(data.grid.affine)
            self.debug("RAS aligned")
            self.debug(self._grid.affine)

            # HACK force a change of focus
            self.set_focus([0, 0, 0, data.nvols], grid=data.grid)
            initial_focus = [float(int(v/2)) for v in data.grid.shape] + [int(data.nvols/2)]
            self.debug("Initial focus (data): %s", initial_focus)
            self.set_focus(initial_focus, grid=data.grid)
            self.debug("Initial focus (std): %s", self._focus)

            # If one of the dimensions has size 1 the data is 2D so
            # maximise the relevant slice. If not, go to standard
            # 3-slice view
            self._toggle_maximise(0, state=0)
            data_axes = data.grid.get_ras_axes()
            for idx in range(3):
                if data.grid.shape[data_axes[idx]] == 1:
                    self._toggle_maximise(idx, state=1)

            self.main_histogram.custom_view = self.main_view_md
            self.main_histogram.qpdata = data
            if data is not None:
                self.main_view_md.cmap_range = data.suggest_cmap_range(vol=int(data.nvols/2), percentile=99)

            self.sig_grid_changed.emit(self._grid)



"""
Quantiphyse - Widgets for data simulation

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets
  
from quantiphyse.gui.options import OptionBox, NumericOption, DataOption, OutputNameOption, ChoiceOption
from quantiphyse.gui.widgets import QpWidget, TitleWidget

from .processes import AddNoiseProcess, SimMotionProcess
    
class AddNoiseWidget(QpWidget):
    """
    Add noise to data
    """
    def __init__(self, **kwargs):
        super(AddNoiseWidget, self).__init__(name="Add noise", icon="noise", 
                                             desc="Add random noise to a data set", 
                                             group="Simulation", **kwargs)

    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self, title="Add Noise", help="noise")
        vbox.addWidget(title)

        self.option_box = OptionBox("Options")
        data = self.option_box.add("Data set", DataOption(self.ivm), key="data")
        self.option_box.add("Gaussian standard deviation", NumericOption(minval=0, maxval=100, default=50), key="std")
        self.option_box.add("Output name", OutputNameOption(src_data=data, suffix="_noisy"), key="output-name")
        vbox.addWidget(self.option_box)

        run_btn = QtGui.QPushButton('Run', self)
        run_btn.clicked.connect(self.run)
        vbox.addWidget(run_btn)
        
        vbox.addStretch(1)
 
    def batch_options(self):
        return "AddNoise", self.option_box.values()
        
    def run(self):
        options = self.batch_options()[1]
        process = AddNoiseProcess(self.ivm)
        process.execute(options)

class SimMotionWidget(QpWidget):
    """
    Widget to simulate random motion on a 4D data set
    """
    def __init__(self, **kwargs):
        super(SimMotionWidget, self).__init__(name="Simulate motion", icon="reg", 
                                              desc="Simulate random motion on a 4D data set", 
                                              group="Simulation", **kwargs)

    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self, title="Simulate Motion", help="sim_motion")
        vbox.addWidget(title)

        self.option_box = OptionBox("Options")
        data = self.option_box.add("Data set", DataOption(self.ivm, include_4d=True, include_3d=False), key="data")
        self.option_box.add("Random translation standard deviation (mm)", NumericOption(minval=0, maxval=5, default=1, decimals=2), key="std")
        self.option_box.add("Random rotation standard deviation (\N{DEGREE SIGN})", NumericOption(minval=0, maxval=10, default=1, decimals=2), key="std_rot")
        self.option_box.add("Padding (mm)", NumericOption(minval=0, maxval=10, default=5, decimals=1), key="padding", checked=True)
        self.option_box.add("Interpolation", ChoiceOption(["Nearest neighbour", "Linear", "Quadratic", "Cubic"], return_values=range(4), default=3), key="order")
        self.option_box.add("Output name", OutputNameOption(src_data=data, suffix="_moving"), key="output-name")
        vbox.addWidget(self.option_box)

        run_btn = QtGui.QPushButton('Run', self)
        run_btn.clicked.connect(self.run)
        vbox.addWidget(run_btn)
        
        vbox.addStretch(1)

    def batch_options(self):
        return "SimMotion", self.option_box.values()
        
    def run(self):
        options = self.batch_options()[1]
        process = SimMotionProcess(self.ivm)
        process.run(options)



"""
Quantiphyse - Functions for loading and querying plugins

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
import os
import glob
import importlib
import logging
import traceback

from quantiphyse.utils.local import get_local_file

PLUGIN_MANIFEST = None
LOG = logging.getLogger(__name__)

def _possible_module(mod_file):
    if os.path.basename(mod_file).startswith("_"):
        return None
    elif os.path.isdir(mod_file): 
        return os.path.basename(mod_file)
    elif mod_file.endswith(".py") or mod_file.endswith(".dll") or mod_file.endswith(".so"):
        return os.path.basename(mod_file).rsplit(".", 1)[0]

def _load_plugins_from_dir(dirname, pkgname, manifest):
    """
    Beginning of plugin system - load modules dynamically from the specified directory

    Then check in module for widgets and/or processes to return
    """
    LOG.debug("Loading plugins from %s", dirname)
    submodules = glob.glob(os.path.join(os.path.abspath(dirname), "*"))
    done = set()
    pythonpath = list(sys.path)
    try:
        sys.path.insert(0, dirname)
        for mod_file in submodules:
            mod = _possible_module(mod_file)
            if mod is not None and mod not in done:
                done.add(mod)
                try:
                    LOG.debug("Trying to import %s", mod)
                    module = importlib.import_module(mod, pkgname)
                    LOG.debug("Got %s (%s)", module.__name__, module.__file__)
                    if hasattr(module, "QP_WIDGETS"):
                        LOG.debug("Widgets found: %s %s", mod, module.QP_WIDGETS)
                        manifest["widgets"] = manifest.get("widgets", []) + module.QP_WIDGETS
                    if hasattr(module, "QP_PROCESSES"):
                        LOG.debug("Processes found: %s %s", mod, module.QP_PROCESSES)
                        manifest["processes"] = manifest.get("processes", []) + module.QP_PROCESSES
                    if hasattr(module, "QP_MANIFEST"):
                        # Module directories are added to the global PYTHONPATH
                        for deps_dir in module.QP_MANIFEST.pop("module-dirs", []):
                            deps_path = os.path.join(dirname, mod_file, deps_dir)
                            if os.path.isdir(deps_path):
                                pythonpath.append(deps_path)
                        # Everything else is added to the global manifest
                        for key, val in module.QP_MANIFEST.items():
                            LOG.debug("%s found: %s %s", key, mod, val)
                            manifest[key] = manifest.get(key, []) + val
                except ImportError:
                    LOG.warn("Error loading plugin: %s", mod)
                    traceback.print_exc()
    finally:
        sys.path = pythonpath

def _load_plugins_from_entry_points(manifest, key="quantiphyse_plugins"):
    import pkg_resources
    for ep in pkg_resources.iter_entry_points(key):
        for key, val in ep.load().items():
            LOG.debug("entry points: found: %s %s", key, val)
            manifest[key] = manifest.get(key, []) + val

def get_plugins(key=None, class_name=None):
    """
    Beginning of plugin system - load widgets dynamically from specified plugins directory
    """
    global PLUGIN_MANIFEST
    if PLUGIN_MANIFEST is None:
        PLUGIN_MANIFEST = {}

        plugin_dirs = {
            "quantiphyse.packages.core" : get_local_file("packages/core"), 
            "quantiphyse.packages.plugins" : get_local_file("packages/plugins"),
        }
        for pkg, plugin_dir in plugin_dirs.items():
            #if os.path.exists(plugin_dir):
            #    __import__(pkg)
            _load_plugins_from_dir(plugin_dir, pkg, PLUGIN_MANIFEST)
        _load_plugins_from_entry_points(PLUGIN_MANIFEST)
    
    if key is not None:
        plugins = PLUGIN_MANIFEST.get(key, [])
        if class_name is not None: 
            plugins = [p for p in plugins if p.__name__ == class_name]
    else:
        plugins = PLUGIN_MANIFEST
    return plugins



"""
Quantiphyse - Custom histogram display widget

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division, unicode_literals, absolute_import, print_function

from matplotlib import cm
import numpy as np

import pyqtgraph as pg

from quantiphyse.data.qpdata import remove_nans

class HistogramWidget(pg.HistogramLUTWidget):
    """
    Displays histogram and colour map from a view metadata object
    and allows them to be updated interactivately

    The histogram is taken from a QpData instance. Normally the 
    colourmap is also taken from the .view property of the QpData, however
    it is possible to set the custom_view property to substitute another
    set of viewing metadata. This is used to implement the 'background'
    main data view.
    """
    def __init__(self, ivl, *args, **kwargs):
        kwargs["fillHistogram"] = False
        super(HistogramWidget, self).__init__(*args, **kwargs)
        self.setBackground(None)

        self._qpdata = None
        self._custom_view = None
        self._vol = 0
        self._updating = False

        ivl.sig_focus_changed.connect(self._focus_changed)
        self.sigLevelChangeFinished.connect(self._levels_changed)
        self.sigLevelsChanged.connect(self._levels_changed)
        self.sigLookupTableChanged.connect(self._lut_changed)

    @property
    def qpdata(self):
        """
        QpData object containing the data to display.

        If the view metadata is not set explicitly, the view property
        of the qpdata is used
        """
        return self._qpdata

    @qpdata.setter
    def qpdata(self, qpdata):
        if self.view is not None:
            try:
                self.view.sig_changed.disconnect(self._view_changed)
            except:
                print("his: except1")

        self._qpdata = qpdata
        if self.view is not None:
            self.view.sig_changed.connect(self._view_changed)
        self._update_cmap()
        self._update_histogram()

    @property
    def view(self):
        """
        Metadata object containing view properties
        """
        if self._qpdata is not None and self._custom_view is None:
            return self._qpdata.view
        else:
            return self._custom_view

    @property
    def custom_view(self):
        return self._custom_view

    @custom_view.setter
    def custom_view(self, view):
        if self.view is not None:
            try:
                self.view.sig_changed.disconnect(self._view_changed)
            except:
                print("his: except2")

        self._custom_view = view
        if self.view is not None:
            self.view.sig_changed.connect(self._view_changed)
        self._update_cmap()

    @property
    def vol(self):
        """
        Current volume index being displayed on the histogram
        """
        return self._vol

    @vol.setter
    def vol(self, vol):
        if vol != self._vol:
            self._vol = vol
            self._update_histogram()

    def _focus_changed(self, focus):
        self.vol = focus[3]

    def _view_changed(self, key, _value):
        if key in ("cmap", "cmap_range"):
            self._update_cmap()

    def _update_histogram(self):
        if self._qpdata is not None:
            arr = remove_nans(self._qpdata.volume(self._vol))
            flat = arr.reshape(-1)
            img = pg.ImageItem(flat.reshape([1, -1]))
            hist = img.getHistogram()
            if hist[0] is None:
                return
            self.plot.setData(*hist)

    def _update_cmap(self):
        if self.view is not None:
            try:
                self._updating = True

                cmap_range = self.view.cmap_range
                if cmap_range is not None:
                    self.region.setRegion(cmap_range)

                cmap = self.view.cmap
                if cmap != "custom":
                    try:
                        self.gradient.loadPreset(cmap)
                    except KeyError:
                        self._set_matplotlib_gradient(cmap)
                else:
                    # FIXME some other components sets a custom cmap
                    # need to convert to ticks and update gradient
                    pass

            finally:
                self._updating = False

    def _levels_changed(self):
        if self.view is not None and not self._updating:
            self.view.cmap_range = list(self.region.getRegion())

    def _lut_changed(self):
        if self.view is not None and not self._updating:
            # FIXME
            pass
            #self.view.lut = [tuple(row) for row in self.gradient.getLookupTable(255)]
            #self.view.cmap = "custom"

    def _set_matplotlib_gradient(self, name):
        """
        Slightly hacky method to copy MatPlotLib gradients to pyqtgraph.

        Is not perfect because Matplotlib specifies gradients in a different way to pyqtgraph
        (specifically there is a separate list of ticks for R, G and B). So we just sample
        the colormap at 10 points which is OK for most slowly varying gradients.
        """
        cmap = getattr(cm, name)
        ticks = [(pos, [255 * v for v in cmap(pos)]) for pos in np.linspace(0, 1, 10)]
        self.gradient.restoreState({'ticks': ticks, 'mode': 'rgb'})

class CurrentDataHistogramWidget(HistogramWidget):
    """
    Displays histogram and colour map from current data
    """
    def __init__(self, ivl, *args, **kwargs):
        super(CurrentDataHistogramWidget, self).__init__(ivl, *args, **kwargs)
        ivl.ivm.sig_current_data.connect(self._current_data_changed)

    def _current_data_changed(self, qpdata):
        self.qpdata = qpdata



"""
Quantiphyse - Module for normalising data, e.g. prior to feature extraction

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division, print_function

import numpy as np

def norm_percentile(data, percentile=90):
    """
    Normalise the data by dividing by a given percentile

    :param data: Numpy array whose last dimension is assumed to be
                 the volume sequence
    """
    norm = np.percentile(data, percentile)
    return data / norm

def norm_median(data, volume_idx=None):
    """
    Normalise the data by dividing by the median of a given volume

    :param data: Numpy array whose last dimension is assumed to be
                 the volume sequence
    """
    if volume_idx is None:
        volume_idx = data.shape[-1] / 2
    return data / np.median(data[:, volume_idx])

def norm_indiv(data):
    """
    Scale each volume individually so it lies between 0 and 1

    :param data: Numpy array whose last dimension is assumed to be
                 the volume sequence
    """
    data = data - np.min(data, axis=0)
    return data / (np.max(data, axis=0) + 0.001)

def norm_sigenh(data, nvols=3):
    """
    Scale each data point by dividing by a 'baseline' value and then subtracting 1

    This results in 'signal enhancement' curves, starting at 0

    :param data: Numpy array whose last dimension is assumed to be
                 the volume sequence
    """
    data_nvols = data.shape[-1]
    tile_shape = [1, ] * data.ndim
    tile_shape[-1] = nvols

    baseline = np.mean(data[..., :min(nvols, data_nvols)], axis=-1)
    baseline_expanded = np.expand_dims(baseline, axis=-1)
    return data / (baseline_expanded + 0.001) - 1

def normalise(data, method, **kwargs):
    """
    Normalise data using named method

    :param data: Numpy array containing data to be normalised
    :param method: One of ``perc``, ``median``, ``indiv``, or ``sigenh``
    :return: Normalised data as matching Numpy array
    """
    norm_methods = {
        "perc" : norm_percentile,
        "median" : norm_median,
        "indiv" : norm_indiv,
        "sigenh" : norm_sigenh,
    }
    if method in norm_methods:
        return norm_methods[method](data, **kwargs)
    else:
        raise ValueError("Unknown normalisation method: %s" % method)



"""
Quantiphyse - radial profile process

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import six

import numpy as np

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.utils import QpException, table_to_extra
from quantiphyse.processes import Process

class RadialProfileProcess(Process):
    """
    Calculate radial profile for a data set
    """
    
    PROCESS_NAME = "RadialProfile"
    
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)
        self.model = QtGui.QStandardItemModel()

    def run(self, options):
        data_items = options.pop('data', None)
        if data_items is None:
            data_items = self.ivm.data.keys()
        elif isinstance(data_items, six.string_types):
            data_items = [data_items,]
            
        if not data_items:
            raise QpException("No data to calculate radial profile")
        
        data_items = [self.ivm.data[name] for name in data_items]
        roi = self.get_roi(options, use_current=False)
        
        #roi_region = options.pop('region', None)
        centre = options.pop('centre')
        if isinstance(centre, six.string_types):
            centre = [float(v) for v in centre.split(",")]
        vol = None
        if len(centre) == 4:
            vol = centre[3]
        
        output_name = options.pop('output-name', "radial-profile")
        bins = options.pop('bins', 20)

        self.model.clear()
        self.rp = {}
        
        grid = data_items[0].grid
        voxel_sizes = grid.spacing

        # Generate an array whose entries are integer values of the distance
        # from the centre. Set masked values to distance of -1
        x, y, z = np.indices(grid.shape)
        r = np.sqrt((voxel_sizes[0]*(x - centre[0]))**2 + (voxel_sizes[1]*(y - centre[1]))**2 + (voxel_sizes[2]*(z - centre[2]))**2)
        if roi is not None:
            roidata = roi.resample(grid).raw()
            r[roidata == 0] = -1
        rmin = r[r > 0].min()

        # Generate histogram of number of voxels in each bin
        # Use the range parameter to ignore masked values with negative distances
        voxels_per_bin, self.edges = np.histogram(r, bins=bins, range=(rmin, r.max()))

        # Prevent divide by zero, if there are no voxels in a bin, this is OK because
        # there will be no data either
        voxels_per_bin[voxels_per_bin == 0] = 1
        self.xvals = [(self.edges[i] + self.edges[i+1])/2 for i in range(len(self.edges)-1)]

        for idx, xval in enumerate(self.xvals):
            self.model.setVerticalHeaderItem(idx, QtGui.QStandardItem(str(xval)))

        for col, data in enumerate(data_items):
            self.model.setHorizontalHeaderItem(col, QtGui.QStandardItem("%s" % data.name))

            if vol is None and data.nvols > 1:
                # All volumes - average over volumes for 4D data
                weights = np.mean(data.resample(grid).raw(), -1)
            else:
                weights = data.resample(grid).volume(vol)

            # Generate histogram by distance, weighted by data
            rpd, _ = np.histogram(r, weights=weights, bins=bins, range=(rmin, r.max()))

            # Divide by number of voxels in each bin to get average value by distance.
            rp = rpd / voxels_per_bin
            for idx, v in enumerate(rp):
                self.model.setItem(idx, col, QtGui.QStandardItem(str(v)))
            self.rp[data.name] = rp

        self.ivm.add_extra(output_name, table_to_extra(self.model, output_name))



"""
Quantiphyse - Utility functions for handling colormaps

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging

import numpy as np

from matplotlib import cm

import pyqtgraph as pg

LOG = logging.getLogger(__name__)

# Standard colour maps. Note that ROI uses jet because it gives more contrast between different indices
CMAPS = ["viridis", "magma", "inferno", "plasma", "cividis", "seismic", "jet", "hot", "gist_heat", "flame", "bipolar", "spectrum", "custom"]
DEFAULT_CMAP = "hot"
DEFAULT_CMAP_ROI = "plasma"

def initial_cmap_range(qpdata, percentile=100):
    """
    Get an initial colourmap range for a data item.

    This is taken by ignoring NaN and infinity and returning
    a percentile of the data range. By default returns
    the full min to max range.

    :return: Sequence of (min, max)
    """
    cmin, cmax = qpdata.range(vol=int(qpdata.nvols/2), percentile=percentile)
    # Issue #101: if min is exactly zero, make it slightly more
    # as a heuristic for data sets where zero=background
    if cmin == 0:
        cmin = 1e-7*cmax

    return cmin, cmax

def get_lut(cmap_name, alpha=255):
    """
    Get the colour lookup table by name.

    Handles Matplotlib as well as pyqtgraph built in colormaps.
    Pyqtgraph is a bit rubbish here - to load a predefined color
    map and extract the lookup table we need to create a
    GradientEditorItem even though we're not doing anything involving
    the GUI.
    """
    if not cmap_name:
        cmap_name = DEFAULT_CMAP

    gradient = pg.GradientEditorItem()
    try:
        gradient.loadPreset(cmap_name)
        LOG.debug("Loaded standard LUT: %s", cmap_name)
    except KeyError:
        gradient.restoreState(get_lut_from_matplotlib(cmap_name))
        LOG.debug("Loaded Matplotlib LUT: %s", cmap_name)

    lut = gradient.getLookupTable(nPts=512, alpha=True)
    LOG.debug("LUT: %s", lut)
    for entry in lut:
        entry[3] = alpha
    return lut

def get_lut_from_matplotlib(cmap_name):
    """
    Slightly hacky method to copy MatPlotLib gradients to pyqtgraph.

    Is not perfect because Matplotlib specifies gradients in a different way to pyqtgraph
    (specifically there is a separate list of ticks for R, G and B). So we just sample
    the colormap at 10 points which is OK for most slowly varying gradients.
    """
    cmap = getattr(cm, cmap_name)
    ticks = [(pos, [255 * v for v in cmap(pos)]) for pos in np.linspace(0, 1, 10)]
    return {'ticks': ticks, 'mode': 'rgb'}

def get_col(lut, idx, range):
    """
    Get RGB color for an index within a range, using a lookup table
    """
    lower, upper = range[0], range[1]
    if lower == upper:
        pos = 0
    else:
        pos = (idx - range[0]) / (range[1] - range[0])
    return lut[int(pos*(len(lut)-1))]

def get_roi_col(roi, region_idx):
    """
    Get RGB color for an ROI region
    """
    lut = get_lut(roi.view.cmap)
    return get_col(lut, region_idx, roi.view.cmap_range)

# Kelly (1965) - set of 20 contrasting colours
# We alter the order a bit to prioritize those that give good contrast to our dark background
# plus we add an 'off white' at the start
KELLY_COLORS = [
    ("off_white", (230, 230, 230)),
    ("vivid_yellow", (255, 179, 0)),
    ("vivid_orange", (255, 104, 0)),
    ("very_light_blue", (166, 189, 215)),
    ("vivid_red", (193, 0, 32)),
    ("grayish_yellow", (206, 162, 98)),
    ("medium_gray", (129, 112, 102)),
    ("strong_purple", (128, 62, 117)),

    # these aren't good for people with defective color vision:
    ("vivid_green", (0, 125, 52)),
    ("strong_purplish_pink", (246, 118, 142)),
    ("strong_blue", (0, 83, 138)),
    ("strong_yellowish_pink", (255, 122, 92)),
    ("strong_violet", (83, 55, 122)),
    ("vivid_orange_yellow", (255, 142, 0)),
    ("strong_purplish_red", (179, 40, 81)),
    ("vivid_greenish_yellow", (244, 200, 0)),
    ("strong_reddish_brown", (127, 24, 13)),
    ("vivid_yellowish_green", (147, 170, 0)),
    ("deep_yellowish_brown", (89, 51, 21)),
    ("vivid_reddish_orange", (241, 58, 19)),
    ("dark_olive_green", (35, 44, 22)),
]

def get_kelly_col(idx, wrap=True):
    """
    Get the Kelly colour for a given index

    :param idx: Index
    :param wrap: If True, wrap index to number of Kelly colours
    :return: (RGB) tuple in range 0-255
    """
    baseidx = idx % len(KELLY_COLORS)
    basecol = KELLY_COLORS[baseidx][1]
    if not wrap:
        raise RuntimeError("Not implemented yet")
    return basecol



"""
Quantiphyse - Widgets for viewing and modifying data orientation and grids

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import math

import numpy as np

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import QpWidget, TitleWidget, NumberGrid
from quantiphyse.gui.options import OptionBox, DataOption, ChoiceOption, OutputNameOption, RunButton, NumericOption, BoolOption
from quantiphyse.utils.enums import Visibility

from .processes import ResampleProcess

class ResampleDataWidget(QpWidget):
    """
    Widget that lets you resample data onto a different grid
    """
    def __init__(self, **kwargs):
        super(ResampleDataWidget, self).__init__(name="Resample Data", icon="resample.png", 
                                                 desc="Resample data onto a different grid",
                                                 group="Utilities", **kwargs)
        
    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        vbox.addWidget(TitleWidget(self))

        self.optbox = OptionBox("Resampling options")
        self.data = self.optbox.add("Data to resample", DataOption(self.ivm), key="data")
        self.resample_type = self.optbox.add("Resampling method", ChoiceOption(["On to grid from another data set", "Upsample", "Downsample"], ["data", "up", "down"]), key="type")
        self.grid_data = self.optbox.add("Use grid from", DataOption(self.ivm), key="grid")
        self.factor = self.optbox.add("Factor", NumericOption(default=2, minval=2, maxval=10, intonly=True), key="factor")
        self.slicewise = self.optbox.add("2D only", BoolOption(), key="2d")
        self.order = self.optbox.add("Interpolation", ChoiceOption(["Nearest neighbour", "Linear", "Quadratic", "Cubic"], [0, 1, 2, 3], default=1), key="order")
        self.output_name = self.optbox.add("Output name", OutputNameOption(src_data=self.data, suffix="_res"), key="output-name")
        vbox.addWidget(self.optbox)
        self.resample_type.sig_changed.connect(self._resample_type_changed)

        self.run = RunButton("Resample", self._run)
        vbox.addWidget(self.run)
        vbox.addStretch(1)

        self._resample_type_changed()

    def _resample_type_changed(self):
        resample_type = self.resample_type.value
        self.optbox.set_visible("grid", resample_type == "data")
        self.optbox.set_visible("factor", resample_type != "data")
        self.optbox.set_visible("order", resample_type != "down")
        self.optbox.set_visible("2d", resample_type != "data")

    def batch_options(self):
        options = self.optbox.values()
        return "Resample", options

    def _run(self):
        _, options = self.batch_options()
        ResampleProcess(self.ivm).run(options)

class OrientDataWidget(QpWidget):
    """
    Widget that lets you tweak the orientation of data
    """
    def __init__(self, **kwargs):
        super(OrientDataWidget, self).__init__(name="Orient Data", icon="inspect.png", 
                                               desc="Manipulate data orientation", 
                                               group="Utilities", **kwargs)
        self._transform_cache = {}
        self.ivm.sig_all_data.connect(self._all_data_changed)

    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self)
        vbox.addWidget(title)

        hbox = QtGui.QHBoxLayout()
        self.options = OptionBox("Re-orient data")
        data = self.options.add("Data item", DataOption(self.ivm), key="data")
        data.sig_changed.connect(self._data_changed)
        self.trans, self.rot = {}, {}
        self.options.add("Translation")
        for axis, label in {2 : "axial", 0 : "sagittal", 1 : "coronal"}.items():
            trans = self.options.add("  %s (mm)" % label.title(), NumericOption(minval=-100, maxval=100, default=0), key="trans-%s" % label)
            trans.sig_changed.connect(self._translate(axis, label))
            self.trans[axis] = trans
        self.options.add("Rotation")
        for axis, label in {2 : "axial", 0 : "sagittal", 1 : "coronal"}.items():
            rot = self.options.add("  %s (degrees)" % label.title(), NumericOption(minval=-180, maxval=180, default=0), key="rot-%s" % label)
            rot.sig_changed.connect(self._rotate(axis, label))
            self.rot[axis] = rot
        hbox.addWidget(self.options)
        vbox.addLayout(hbox)

        self.gridview = GridView(self.ivm, self.ivl)
        vbox.addWidget(self.gridview)

        hbox = QtGui.QHBoxLayout()
        reset_btn = QtGui.QPushButton("Reset to original")
        reset_btn.clicked.connect(self._reset)
        hbox.addWidget(reset_btn)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        vbox.addStretch(1) 

    def activate(self):
        self._data_changed()

    def _all_data_changed(self, data):
        for name in list(self._transform_cache.keys()):
            if name not in data:
                del self._transform_cache[name]

    def _data_changed(self):
        name = self.options.values()["data"]
        qpdata = self.ivm.data.get(name, self.ivm.rois.get(name, None))
        self.gridview.set_data(qpdata)
        if qpdata is not None:
            if name not in self._transform_cache:
                self._transform_cache[name] = ([0, 0, 0], [0, 0, 0])
            translation, rotations = self._transform_cache[name]
            for axis in range(3):
                self.trans[axis].value = translation[axis]
                self.rot[axis].value = rotations[axis]
            self._set()

    def _translate(self, axis, label):
        def _trans():
            name = self.gridview.data.name
            trans = self.options.values()["trans-%s" % label]
            self._transform_cache[name][0][axis] = trans
            self._set()
        return _trans

    def _rotate(self, axis, label):
        def _rot():
            name = self.gridview.data.name
            angle = self.options.values()["rot-%s" % label]
            if axis == 1: angle = -angle
            self._transform_cache[name][1][axis] = angle
            self._set()
        return _rot

    def _reset(self):
        name = self.gridview.data.name
        del self._transform_cache[name]
        self._data_changed()

    def _set(self):
        name = self.gridview.data.name
        affine = self.gridview.data.grid.affine_orig
        grid_centre = [float(dim) / 2 for dim in self.gridview.data.grid.shape]
        world_centre = np.dot(affine[:3, :3], grid_centre)
        self.debug("Initial affine\n%s", affine)
        translation, rotations = self._transform_cache[name]

        R = np.identity(3)
        for axis in range(3):
            angle = rotations[axis]
            rot3d = self._rotmtx_3d(axis, angle)
            affine[:3, :3] = np.dot(rot3d, affine[:3, :3])
            R = np.dot(rot3d, R)

        origin_offset = world_centre - np.dot(R, world_centre)
        origin_offset += translation
        self.debug("Origin offset\n%s", origin_offset)
        affine[:3, 3] += origin_offset

        self.debug("Final affine\n%s", affine)
        self.gridview.data.grid.affine = affine
        self.gridview.update()
        if self.gridview.data == self.ivm.main:
            self.ivm.sig_main_data.emit(self.ivm.main)
        if self.gridview.data.view.visible == Visibility.SHOW or self.gridview.data == self.ivm.main:
            self.ivl.redraw()

    def _rotmtx_3d(self, axis, angle):
        # FIXME this is not quite right when rotating in a plane where
        # the basis vectors have different lengths
        c, s = math.cos(math.radians(angle)), math.sin(math.radians(angle))
        rot2d = np.array([[c, -s], [s, c]])
        rot3d = np.identity(3)
        if axis == 0:
            rot3d[1:, 1:] = rot2d
        elif axis == 1:
            rot3d[0, 0] = rot2d[0, 0]
            rot3d[0, 2] = rot2d[0, 1]
            rot3d[2, 0] = rot2d[1, 0]
            rot3d[2, 2] = rot2d[1, 1]
        elif axis == 2:
            rot3d[:2, :2] = rot2d
        self.debug("3d rotation matrix: %i %f", axis, angle)
        self.debug("\n%s", rot3d)
        return rot3d

class GridView(QtGui.QWidget):

    COORD_LABELS = {
        0 : "unknown",
        1 : "scanner",
        2 : "aligned",
        3 : "Talairach",
        4 : "MNI",
    }

    def __init__(self, ivm, ivl, readonly=False):
        self.ivm, self.ivl = ivm, ivl
        self.data = None
        self._updating = False

        QtGui.QWidget.__init__(self)
        grid = QtGui.QGridLayout()
        self.setLayout(grid)

        grid.addWidget(QtGui.QLabel("Grid->World Transform"), 1, 0)
        self.transform = NumberGrid(initial=np.identity(3), expandable=(False, False), fix_height=True, fix_width=True, readonly=readonly)
        self.transform.sig_changed.connect(self._changed)
        grid.addWidget(self.transform, 2, 0)

        grid.addWidget(QtGui.QLabel("Origin"), 1, 1)
        self.origin = NumberGrid(initial=[[0],]*3, expandable=(False, False), fix_height=True, fix_width=True, readonly=readonly)
        self.origin.sig_changed.connect(self._changed)
        grid.addWidget(self.origin, 2, 1)

        grid.addWidget(QtGui.QLabel("Co-ordinate system: "), 3, 0)
        self.coord_label = QtGui.QLabel("unknown")
        grid.addWidget(self.coord_label, 3, 1)

        grid.setColumnStretch(3, 1)
    
    def set_data(self, data):
        if data == self.data:
            return
        self.data = data
        self.update()

    def update(self):
        self._updating = True
        try:
            if self.data is not None:
                if hasattr(self.data, "nifti_header"):
                    self.coord_label.setText(self.COORD_LABELS[int(self.data.nifti_header['sform_code'])])
                self.transform.setValues(self.data.grid.transform)
                self.origin.setValues([[x,] for x in self.data.grid.origin])
            else:
                self.coord_label.setText("unknown")
                self.transform.setValues(np.identity(3))
                self.origin.setValues([[0],]*3)
        finally:
            self._updating = False

    def _changed(self):
        if self.data is not None and not self._updating:
            affine = self.data.grid.affine
            if self.transform.valid():
                affine[:3, :3] = self.transform.values()
            if self.origin.valid():
                affine[:3, 3] = [x[0] for x in self.origin.values()]
            self.data.grid.affine = affine
            # HACK
            if self.data == self.ivm.current_data:
                self.ivm.sig_current_data.emit(self.ivm.current_data)
            elif self.data == self.ivm.main:
                self.ivm.sig_main_data.emit(self.ivm.main)
                


"""
Quantiphyse - Widgets which supports the writing of batch files

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import QpWidget, TitleWidget, RunBox
from quantiphyse.utils.batch import BatchScript

class BatchBuilderWidget(QpWidget):
    """
    Simple widget for building and running batch files
    """
    def __init__(self, **kwargs):
        super(BatchBuilderWidget, self).__init__(name="Batch Builder", icon="batch", 
                                                 desc="Simple helper for building and running batch files", 
                                                 group="Utilities", **kwargs)
        self.process = BatchScript(self.ivm, stdout=open(os.devnull, "w"), quit_on_exit=False)

    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self, title="Batch Builder", subtitle=self.description, batch_btn=False)
        vbox.addWidget(title)

        self.proc_edit = QtGui.QPlainTextEdit()
        self.proc_edit.textChanged.connect(self._validate)
        vbox.addWidget(self.proc_edit, stretch=2)

        hbox = QtGui.QHBoxLayout()
        #self.new_case_btn = QtGui.QPushButton("New Case")
        #hbox.addWidget(self.new_case_btn)
        #hbox.setAlignment(self.new_case_btn, QtCore.Qt.AlignTop)

        self.reset_btn = QtGui.QPushButton("Reset")
        self.reset_btn.clicked.connect(self._reset)
        hbox.addWidget(self.reset_btn)
        hbox.addStretch(1)

        self.save_btn = QtGui.QPushButton("Save")
        self.save_btn.clicked.connect(self._save)
        hbox.addWidget(self.save_btn)
        vbox.addLayout(hbox)

        self.run_box = RunBox(self._get_process, self._get_options)
        vbox.addWidget(self.run_box)

        self.proc_warn = QtGui.QLabel("")
        self.proc_warn.setStyleSheet("QLabel { color : red; }")
        vbox.addWidget(self.proc_warn)

        self.default_dir = os.getcwd()
        self.changed = False
        self._reset()

    def activate(self):
        self.ivm.sig_main_data.connect(self._update)
        self.ivm.sig_all_data.connect(self._update)
        self._update()

    def deactivate(self):
        self.ivm.sig_main_data.disconnect(self._update)
        self.ivm.sig_all_data.disconnect(self._update)

    def _update(self):
        if not self.changed:
            self._reset()

    def _reset(self):
        if self.changed and self.proc_edit.toPlainText().strip() != "":
            btn = QtGui.QMessageBox.warning(self, "Reset to current data",
                                            "Changes to batch file will be lost",
                                            QtGui.QMessageBox.Ok | QtGui.QMessageBox.Cancel)
            if btn != QtGui.QMessageBox.Ok:
                return
        
        t = ""
        if self.ivm.main is not None:
            filedata = [d for d in self.ivm.data.values() if hasattr(d, "fname") and d.fname is not None]
            filerois = [d for d in self.ivm.rois.values() if hasattr(d, "fname") and d.fname is not None]
            
            casedir = os.getcwd()
            if hasattr(self.ivm.main, "fname") and self.ivm.main.fname is not None:
                casedir = os.path.dirname(self.ivm.main.fname)

            t += "OutputFolder: qp_out\n"
            t += "Debug: False\n"
            t += "\n"
            t += "Processing:\n"
            t += "  - Load:\n"
            if filedata:
                t += "      data:\n"
                for qpd in filedata:
                    d, f = os.path.split(qpd.fname)
                    if d == casedir:
                        path = f
                    else:
                        path = qpd.fname
                    t += "        %s: %s\n" % (path, qpd.name)
            if filerois:
                t += "      rois:\n"
                for qpd in filerois:
                    d, f = os.path.split(qpd.fname)
                    if d == casedir:
                        path = f
                    else:
                        path = qpd.fname
                    t += "        %s: %s\n" % (path, qpd.name)
            t += "\n"
            t += "  # Additional processing steps go here\n"
            t += "\n"
            t += "  - SaveAllExcept:\n"
            for d in filedata:
                t += "      %s:\n" % d.name
            for d in filerois:
                t += "      %s:\n" % d.name
            t += "\n"

            t += "Cases:\n"
            t += "  Case1:\n"
            t += "    Folder: %s\n" % casedir
            t += "\n"

        self.proc_edit.setPlainText(t)
        self.changed = False
        self.reset_btn.setEnabled(False)

    def _validate(self):
        self.changed = True
        self.reset_btn.setEnabled(True)
        t = self.proc_edit.toPlainText()
        self.proc_warn.setText("")
        if '\t' in self.proc_edit.toPlainText():
            self.proc_warn.setText("Tabs detected")
        else:
            try:
                self.process.run({"yaml" : t, "mode" : "check"})
            except Exception as e:
                self.proc_warn.setText("Invalid script: %s" % str(e))

    def _get_process(self):
        return self.process

    def _get_options(self):
        return {"yaml" : self.proc_edit.toPlainText()}

    def _save(self):
        fname, _ = QtGui.QFileDialog.getSaveFileName(self, 'Save batch file', 
                                                     dir=self.default_dir, 
                                                     filter="YAML files (*.yaml)")
        if fname != '':
            f = open(fname, "w")
            try:
                f.write(self.proc_edit.toPlainText())
            finally:
                f.close()
        else: # Cancelled
            pass



"""
Quantiphyse - Plotting widgets

Copyright (c) 2013-2020 University of Oxford

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import division, unicode_literals, absolute_import

import numpy as np
import pyqtgraph as pg
try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from scipy.interpolate import UnivariateSpline

from quantiphyse.gui.widgets import OptionsButton
from quantiphyse.gui.colors import get_kelly_col
from quantiphyse.utils import LogSource, get_icon, norecurse

class PlotOptions(QtGui.QDialog):
    """
    Options dialog for a line plot
    """

    sig_options_changed = QtCore.Signal(object)

    def __init__(self, graph=None, **kwargs):
        QtGui.QDialog.__init__(self, graph)
        self.graph = graph
        self.sig_enh = False
        self.smooth = False

        self.setWindowTitle('Plot options')
        grid = QtGui.QGridLayout()
        self.setLayout(grid)

        row = 0
        # Display mode
        if kwargs.get("display_mode", True):
            grid.addWidget(QtGui.QLabel("Display mode"), row, 0)
            mode_combo = QtGui.QComboBox()
            mode_combo.addItem("Signal")
            mode_combo.addItem("Signal Enhancement")
            mode_combo.currentIndexChanged.connect(self._mode_changed)
            grid.addWidget(mode_combo, row, 1)
            row += 1

            # Signal enhancement baseline
            self.se_lbl = QtGui.QLabel('Signal enhancement: Use first')
            self.se_lbl.setEnabled(False)
            grid.addWidget(self.se_lbl, row, 0)

            hbox = QtGui.QHBoxLayout()
            self._norm_frames_spin = QtGui.QSpinBox()
            self._norm_frames_spin.setValue(3)
            self._norm_frames_spin.setMinimum(1)
            self._norm_frames_spin.setMaximum(100)
            self._norm_frames_spin.valueChanged.connect(self._changed)
            self._norm_frames_spin.setEnabled(False)
            hbox.addWidget(self._norm_frames_spin)
            self.se_lbl2 = QtGui.QLabel('frames as baseline')
            self.se_lbl2.setEnabled(False)
            hbox.addWidget(self.se_lbl2)
            hbox.addStretch(1)
            grid.addLayout(hbox, row, 1)
            row += 1

        # Y-axis scale
        if kwargs.get("y_scale", True):
            hbox = QtGui.QHBoxLayout()
            self.auto_y_cb = QtGui.QCheckBox('Automatic Y axis scale', self)
            self.auto_y_cb.setChecked(True)
            self.auto_y_cb.stateChanged.connect(self._auto_y_changed)
            grid.addWidget(self.auto_y_cb, row, 0)

            self.min_lbl = QtGui.QLabel("Min")
            self.min_lbl.setEnabled(False)
            hbox.addWidget(self.min_lbl)
            self.min_spin = QtGui.QDoubleSpinBox()
            self.min_spin.setMinimum(-1e20)
            self.min_spin.setMaximum(1e20)
            self.min_spin.valueChanged.connect(self._changed)
            self.min_spin.setEnabled(False)
            hbox.addWidget(self.min_spin)

            self.max_lbl = QtGui.QLabel("Max")
            self.max_lbl.setEnabled(False)
            hbox.addWidget(self.max_lbl)
            self.max_spin = QtGui.QDoubleSpinBox()
            self.max_spin.setMinimum(-1e20)
            self.max_spin.setMaximum(1e20)
            self.max_spin.valueChanged.connect(self._changed)
            self.max_spin.setEnabled(False)
            hbox.addWidget(self.max_spin)

            hbox.addStretch(1)
            grid.addLayout(hbox, row, 1)
            row += 1

        # Smoothing
        if kwargs.get("smoothing", True):
            smooth_cb = QtGui.QCheckBox('Smooth curves', self)
            smooth_cb.setChecked(self.smooth)
            smooth_cb.stateChanged.connect(self._smooth_changed)
            grid.addWidget(smooth_cb, row, 0)
            row += 1

    @property
    def yrange(self):
        return self.min_spin.value(), self.max_spin.value()

    @yrange.setter
    def yrange(self, yrange):
        try:
            self.min_spin.blockSignals(True)
            self.max_spin.blockSignals(True)
            self.min_spin.setValue(yrange[0])
            self.max_spin.setValue(yrange[1])
            self.sig_options_changed.emit(self)
        finally:
            self.min_spin.blockSignals(False)
            self.max_spin.blockSignals(False)

    @property
    def autorange(self):
        return self.auto_y_cb.isChecked()

    @autorange.setter
    def autorange(self, autorange):
        self.auto_y_cb.setChecked(autorange)

    @property
    def norm_frames(self):
        return self._norm_frames_spin.value()

    def _mode_changed(self, idx):
        self.sig_enh = (idx == 1)
        self.se_lbl.setEnabled(self.sig_enh)
        self._norm_frames_spin.setEnabled(self.sig_enh)
        self.se_lbl2.setEnabled(self.sig_enh)
        self.sig_options_changed.emit(self)

    def _auto_y_changed(self, selected):
        self.min_lbl.setEnabled(not selected)
        self.min_spin.setEnabled(not selected)
        self.max_lbl.setEnabled(not selected)
        self.max_spin.setEnabled(not selected)
        self.sig_options_changed.emit(self)
        
    def _smooth_changed(self, state):
        self.smooth = state
        self.sig_options_changed.emit(self)

    def _changed(self):
        self.sig_options_changed.emit(self)

class LinePlot(LogSource):
    """
    A 1-D array of data to be plotted as a line
    """
    def __init__(self, graph, yvalues, xvalues=None, axes=None, name=None, **kwargs):
        LogSource.__init__(self)
        if axes is None:
            self.axes = graph.axes
        if xvalues is None:
            xvalues = range(len(yvalues))

        self.name = name
        self.graph = graph
        self.axes = axes
        
        self.graph.options.sig_options_changed.connect(self._options_changed)

        self.yvalues = np.copy(np.array(yvalues, dtype=np.double))
        self.xvalues = np.copy(np.array(xvalues, dtype=np.double))
        self.line_col = kwargs.get("line_col", (255, 255, 255))
        self.line_width = kwargs.get("line_width", 1.0)
        self.line_style = kwargs.get("line_style", QtCore.Qt.SolidLine)
        self.point_col = kwargs.get("point_col", "k")
        self.point_size = kwargs.get("point_size", 5.0)
        self.point_brush = kwargs.get("point_brush", (200, 200, 200))

        self.legend_item = None
        self.graphics_items = ()
        self.show()

    def show(self):
        """
        Draw the line on to the plot
        """
        if self.graph.options.sig_enh:
            norm_frames = min(len(self.yvalues), self.graph.options.norm_frames)
            mean = np.mean(self.yvalues[:norm_frames])
            plot_values = self.yvalues / mean - 1
        else:
            plot_values = self.yvalues

        if self.graph.options.smooth:
            indexes = range(len(plot_values))
            # Tolerance does not scale by data value to scale input
            spline = UnivariateSpline(indexes, plot_values/plot_values.max(), s=0.1, k=4)
            self.debug("Number of knots in B-spline smoothing: %i", len(spline.get_knots()))
            line_values = spline(indexes)*plot_values.max()
        else:
            line_values = plot_values

        self.hide()
        # Plot line and points separately as line may be smoothed
        line = self.axes.plot(self.xvalues, line_values, 
                              pen=pg.mkPen(color=self.line_col, width=self.line_width, style=self.line_style))

        pts = self.axes.plot(self.xvalues, plot_values, 
                             pen=None, 
                             symbolBrush=self.point_brush, 
                             symbolPen=self.point_col,
                             symbolSize=self.point_size)

        self.legend_item = line
        self.graphics_items = (line, pts)

    def hide(self):
        """
        Remove the line from the plot
        """
        items = self.graphics_items
        self.graphics_items = ()
        self.legend_item = None

        for item in items:
            self.axes.removeItem(item)

    def _options_changed(self):
        if self.graphics_items:
            self.show()

class Plot(QtGui.QWidget):
    """
    Widget for plotting graphs
    """
    def __init__(self, parent=None, title="", **kwargs):
        """
        Create a plot widget

        :param qpo: Global options
        """
        QtGui.QWidget.__init__(self, parent)
        self.items = []
        self.updating = False
        self.legend_pos = (30, 30)

        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        hbox = QtGui.QHBoxLayout()
        hbox.addStretch(1)

        if kwargs.get("opts_btn", True):
            self.opts_btn = OptionsButton()
            self.opts_btn.clicked.connect(self._show_options)
            hbox.addWidget(self.opts_btn)


        if kwargs.get("clear_btn", False):
            clear_icon = QtGui.QIcon(get_icon("clear"))
            self.clear_btn = QtGui.QPushButton(self)
            self.clear_btn.setIcon(clear_icon)
            self.clear_btn.setIconSize(QtCore.QSize(14, 14))
            self.clear_btn.setToolTip("Clear curves")
            self.clear_btn.clicked.connect(self.clear)
            hbox.addWidget(self.clear_btn)

        vbox.addLayout(hbox)

        self.graphics_layout = pg.GraphicsLayoutWidget()
        self.graphics_layout.setBackground(background=None)
        vbox.addWidget(self.graphics_layout, 10)

        self.axes = self.graphics_layout.addPlot()
        self.axes.setTitle(title)
        self.axes.sigRangeChanged.connect(self._range_changed)
        self.axes.vb.sigRangeChangedManually.connect(self._range_changed_manually)

        if kwargs.get("twoaxis", False):
            # For a second y-axis, create a new ViewBox, link the right axis to its coordinate system
            self.axes_alt = pg.ViewBox()
            self.axes.scene().addItem(self.axes_alt)
            self.axes.getAxis('right').linkToView(self.axes_alt)
            self.axes_alt.setXLink(self.axes)
            self.axes.vb.sigResized.connect(self._update_plot_viewbox)
        else:
            self.axes_alt = None

        self.options = PlotOptions(self, **kwargs)
        self._options_changed()
        self.options.sig_options_changed.connect(self._options_changed)
        self._regenerate_legend()

    def set_xlabel(self, name):
        """
        Set the label for the X axis
        """
        self.axes.setLabel('bottom', name)

    def set_ylabel(self, name):
        """
        Set the label for the Y axis
        """
        self.axes.setLabel('left', name)

    def add_line(self, yvalues, name=None, xvalues=None, **kwargs):
        """
        Add a line to the plot

        :param values: 1-D Array of data values
        """
        if "line_col" not in kwargs:
            kwargs["line_col"] = get_kelly_col(len(self.items))

        axes = kwargs.get("axes", "main") 
        if axes.lower() == "main":
            axes = self.axes
        elif axes.lower() == "alt":
            axes = self.axes_alt
        else:
            raise ValueError("Unknown axes: %s" % axes)

        line = LinePlot(self, yvalues, xvalues=xvalues, name=name, axes=axes, **kwargs)
        self.items.append(line)
        self._contents_changed()
        return line

    def remove(self, item):
        """
        Remove an item from the plot
        """
        item.hide()
        self.items.remove(item)
        self._contents_changed()

    def clear(self):
        """
        Clear the plot
        """
        for item in self.items:
            item.hide()
        self.items = []
        self._contents_changed()

    def _show_options(self):
        """
        Show the options dialog box
        """
        self.options.show()
        self.options.raise_()

    def _update_plot_viewbox(self):
        # Required to keep the right and left axis plots in sync with each other
        self.axes_alt.setGeometry(self.axes.vb.sceneBoundingRect())
        
        # Need to re-update linked axes since this was called
        # incorrectly while views had different shapes.
        # (probably this should be handled in ViewBox.resizeEvent)
        self.axes_alt.linkedViewChanged(self.axes.vb, self.axes_alt.XAxis)

    @norecurse
    def _range_changed(self, *args):
        self.options.yrange = self.axes.viewRange()[1]

    @norecurse
    def _range_changed_manually(self, *args):
        self.options.yrange = self.axes.viewRange()[1]
        self.options.autorange = False

    def _contents_changed(self):
        self._regenerate_legend()

    @norecurse
    def _options_changed(self, *args):
        if self.axes_alt is not None:
            self.axes_alt.enableAutoRange() # FIXME
        if self.options.autorange:
            self.axes.enableAutoRange()
        else: 
            self.axes.disableAutoRange()
            self.axes.setYRange(*self.options.yrange)

    def _regenerate_legend(self):
        # Replaces any existing legend but keep position the same in case user moved it
        # This is necessary because of what seems to be a bug in pyqtgraph which prevents
        # legend items from being removed
        if self.axes.legend:
            self.legend_pos = self.axes.legend.pos()
            if self.axes.legend.scene(): 
                self.axes.legend.scene().removeItem(self.axes.legend)
            self.axes.legend = None
    
        for item in self.items:
            if item.name:
                if not self.axes.legend:
                    self.axes.addLegend(offset=self.legend_pos)
                self.axes.legend.addItem(item.legend_item, item.name)
